/* $Id: mn.c,v 1.239 2001/10/20 14:44:19 jm Exp $
 * Mobile Node - state machine and main loop
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/icmp.h>
//#include <linux/if.h>

#include "agentapi.h"
#include "debug.h"
#include "util.h"
#include "dyn_ip.h"
#include "mn.h"
#ifndef DYN_TARGET_WINDOWS
#include "monitor.h"
#endif /* DYN_TARGET_WINDOWS */
#include "fixed_fd_zero.h"
#ifdef INCLUDE_IPAY
#include "mn_ipay.h"
#endif

extern int real_tunnel_up; 			/* from mn_tunnel */
extern int mn_traces;
struct mn_data mn;
struct mn_config config; 			/* configurated values */
struct timeval timers[TIMER_COUNT]; /* Ending times for the timers */
char *program_name;
int mn_traces = 0;

extern struct nw_iface usable_nw_ifaces[NW_IFACE_ARRAY_LEN];
extern int usable_nw_iface_count;

/* Route Optimization Cache contains an entry for each 
 * known CR ( == a peer MR). */
struct ro_entry ro_cache[RO_CACHE_MAX_SIZE];
int ro_entry_count = 0;

int get_ro_entry_count() {
	return ro_entry_count;
}

/* handle_ro_cache_triggers_and_sockets: Check if HoTI + CoTI or Registration Reply 
 * messages need to be sent to CRs. Put also the RO Caches listener sockets to the 
 * fr_set.
 */
int handle_ro_cache_triggers_and_sockets(fd_set *set)
{		
	int i, j;
	char next_iface[IFNAMSIZ];

	for (i = 0; i < ro_entry_count; i++) 
	{	
		/* Check if HoTI+CoTI need to be resent because of no replies received */
		if (ro_cache[i].send_init_msgs == 0 &&
			timerisset(&ro_cache[i].init_send_time) && 
			(time(NULL) - ro_cache[i].init_send_time.tv_sec >= RESEND_HOTI_COTI_TIME))
		{
			print_trace(mn_traces, "No HoT or CoT received from CR %s in %d sec" 
				"-> send HoTI+CoTI through next iface\n", inet_ntoa(ro_cache[i].cr_hoa), 
				RESEND_HOTI_COTI_TIME);
			
			bind_cr_reg_connection_to_device(&ro_cache[i], ro_cache[i].last_coti_iface);
			get_next_usable_iface(ro_cache[i].last_coti_iface, next_iface);
			bind_socket_to_device(ro_cache[i].coti_socket, next_iface);
			memcpy(ro_cache[i].last_coti_iface, next_iface, IFNAMSIZ);
			go_to_resend_hoti_coti_through_new_iface_state(&ro_cache[i]);
			
			//ro_cache[i].send_init_msgs = 1;
		}
		
		DEBUG(DEBUG_TIMERS, "CR %s Keepalive timer set to %ld.%ld\n", 
			inet_ntoa(ro_cache[i].cr_hoa), ro_cache[i].keepalive_timer.tv_sec, 
				ro_cache[i].keepalive_timer.tv_usec);
				
		/* Check if HoTI+CoTI or RegRep needs to be sent */
		if (ro_cache[i].send_init_msgs) 
		{
			ro_cache[i].send_init_msgs = 0;
			send_hoti_coti(&ro_cache[i]);
		} 
		if (ro_cache[i].send_regrep) {
			send_regrep_to_cr(&ro_cache[i]);
		}
		
		/* TODO: Check if some MR CoA's KRm key is expiring */

		/* Put RO Entry's sockets into the fd_set */
		FD_SET(ro_cache[i].registration_socket, set);
		FD_SET(ro_cache[i].coti_socket, set);
		DEBUG(DEBUG_INFO, "ro_cache[%d]: HoA %s, ", i, inet_ntoa(ro_cache[i].cr_hoa));
		DEBUG(DEBUG_INFO, "TNL dst %s, TNL role MR %d\n", inet_ntoa(ro_cache[i].tnl_dst),
			ro_cache[i].tunneling_as_mr);
		
		/* Print CR's RTT array */
	  	DEBUG(DEBUG_INFO, "RTTs to CR: [ ");
	  	for (j = 0; j < RTT_ARRAY_LEN; j++) {
			DEBUG(DEBUG_INFO, "%d ", ro_cache[i].RTTs[j]);
	  	}
		DEBUG(DEBUG_INFO, "]\n");
	}
	return 0;
}

/* cp_iface_list_into_ro_entry: Copy MR CoAs from the usable iface list.
 */
int cp_iface_list_into_ro_entry(struct ro_entry *cr)
{
	int i;
	
	for (i = 0; i < usable_nw_iface_count; i++)
		cr->mr_coas[i].coa.s_addr = usable_nw_ifaces[i].ip_addr.s_addr;
	
	cr->mr_coa_count = usable_nw_iface_count;
	
	return 0;
}

/* start_using_krm_of_iface_if_valid: Check if RO Entry has a valid KRm for
 * the device. Start using that KRm if it's ok.
 * 
 * Returns: 1 if a valid KRm for the iface found, 0 if not
 */
int start_using_krm_of_iface_if_valid(struct ro_entry *cr, char *iface_name)
{
	struct in_addr dev_addr;
	int i;
	
	DEBUG(DEBUG_INFO, "start_using_krm_of_iface_if_valid: CR %s, iface %s\n",
		inet_ntoa(cr->cr_hoa), iface_name);
	
	get_iface_address(iface_name, &dev_addr);
	
	/* If the new iface is not yet in the mr_coas list, add it there */
	for (i = 0; i < cr->mr_coa_count; i++)
	{
		if (dev_addr.s_addr == cr->mr_coas[i].coa.s_addr)	{
			if (cr->mr_coas[i].krm[0] != 0 || cr->mr_coas[i].krm[1] != 0) {
				cr->hot_nonce_index = cr->mr_coas[i].hot_nonce_index;
				cr->cot_nonce_index = cr->mr_coas[i].cot_nonce_index;
				cr->krm[0] 			= cr->mr_coas[i].krm[0];
				cr->krm[1] 			= cr->mr_coas[i].krm[1];
				DEBUG(DEBUG_INFO, " - KRm found -> start using this: %llu %llu\n",
					cr->krm[0], cr->krm[1]);
				return 1;
			} else {
				DEBUG(DEBUG_INFO, " - No KRm found\n");
				return 0;
			}
		} else if (i == cr->mr_coa_count - 1) {
			DEBUG(DEBUG_INFO, " - Device not found from the mr_coas array "
				"-> add it there with an empty KRm\n");
			cr->mr_coas[cr->mr_coa_count].coa.s_addr = cr->mr_coa.s_addr;
			cr->krm[0] = 0;
			cr->krm[1] = 0;
			cr->mr_coa_count++;
			return 0;
		}
	}
	return 0;
}
/* bind_cr_reg_connection_to_device: Bind registration socket to the new nw 
 * iface dev_name. Update fields device_in_use, mr_coa and coa_mip_socket. 
 * Add the mr_coa into the mr_coas array if it's not already there.
 */
int bind_cr_reg_connection_to_device(struct ro_entry *cr, char *dev_name) 
{
	int i;
	DEBUG(DEBUG_INFO, "Bind CR %s Registration connection to nw iface %s\n", 
		inet_ntoa(cr->cr_hoa), dev_name);
	memcpy(cr->device_in_use, dev_name, IFNAMSIZ);
	bind_socket_to_device(cr->registration_socket, dev_name);
	get_iface_address(dev_name, &cr->mr_coa);
	cr->coa_mip_socket = get_iface_mip_socket(dev_name);
	
	/* If the new iface is not yet in the mr_coas list, add it there */
	for (i = 0; i < cr->mr_coa_count; i++)
	{
		if (cr->mr_coa.s_addr == cr->mr_coas[i].coa.s_addr)	{
			break; 
		} else if (i == cr->mr_coa_count - 1) {
			DEBUG(DEBUG_INFO, " - The new MR CoA %s not found from the mr_coas array "
				"-> add it there\n", inet_ntoa(cr->mr_coa));
			cr->mr_coas[cr->mr_coa_count].coa.s_addr = cr->mr_coa.s_addr;
			cr->krm[0] = 0;
			cr->krm[1] = 0;
			cr->mr_coa_count++;
		}
	}
	DEBUG(DEBUG_INFO, " - IP address %s\n", inet_ntoa(cr->mr_coa));
	DEBUG(DEBUG_INFO, " - MIP listener socket fd %d\n", cr->coa_mip_socket);

	return 1;
}

/* bind_cr_connection_to_device: Update all fields in the RO Entry for the new nw iface.
 * The new device has to be verified to be usable for HAaRO before calling this function!
 * 
 */
int bind_cr_connection_to_device(struct ro_entry *cr, char *dev_name) {
	int i;

	DEBUG(DEBUG_INFO, "bind_cr_connection_to_device: CR %s, device %s\n", inet_ntoa(cr->cr_hoa), dev_name);
	
	memcpy(cr->device_in_use, dev_name, IFNAMSIZ);
	memcpy(cr->last_coti_iface, dev_name, IFNAMSIZ);
	bind_socket_to_device(cr->registration_socket, dev_name);
	bind_socket_to_device(cr->coti_socket, dev_name);
	get_iface_address(dev_name, &cr->mr_coa);
	cr->coa_mip_socket = get_iface_mip_socket(dev_name);
	
	DEBUG(DEBUG_INFO, " - new MR CoA %s\n", inet_ntoa(cr->mr_coa));
	DEBUG(DEBUG_INFO, " - MIP listener socket fd %d\n", cr->coa_mip_socket);
		
	/* Find the iface from the mr_coas list to get the krm */
	for (i = 0; i < cr->mr_coa_count; i++)
	{
		if (cr->mr_coa.s_addr == cr->mr_coas[i].coa.s_addr)	
		{
			cr->hot_nonce_index = cr->mr_coas[i].hot_nonce_index;
			cr->cot_nonce_index = cr->mr_coas[i].cot_nonce_index;
			cr->krm[0] 			= cr->mr_coas[i].krm[0];
			cr->krm[1] 			= cr->mr_coas[i].krm[1];
			DEBUG(DEBUG_INFO, " - KRm %llu %llu\n", cr->krm[0], cr->krm[1]);
			break;
		} 
		if (i == cr->mr_coa_count - 1) 
		{
			/* Iface not found from the array */
			DEBUG(DEBUG_INFO, "The new MR CoA %s not found from "
				"the mr_coas array -> add it there\n", inet_ntoa(cr->mr_coa));
			cr->mr_coas[cr->mr_coa_count].coa.s_addr = cr->mr_coa.s_addr;
			cr->krm[0] = 0;
			cr->krm[1] = 0;
			cr->mr_coa_count++;
		}
	}
	return 1;
}

/* get_new_ro_entry: Create a new entry in the RO Cache and open 
 * the socket in this entry.
 * 
 * Returns: a pointer to the new entry. 
 */
struct ro_entry *get_new_ro_entry() {
	int one = 1;
	struct ro_entry *new_entry;
	
	DEBUG(DEBUG_INFO, "New RO Cache Entry: ro_cache[%d]\n",
		ro_entry_count);
	new_entry = &ro_cache[ro_entry_count];
	ro_entry_count++;
	
	/* Init counters etc. */
	new_entry->krm[0] 					= 0;
	new_entry->krm[1] 					= 0; 
	new_entry->cr_coa_count 			= 0;
	new_entry->cr_nw_count				= 0;
	new_entry->tnl_count				= 0;
	new_entry->connection_initialized 	= 0;
	new_entry->rr_state 				= RR_INACTIVE;
	new_entry->keepalive_timer.tv_sec 	= 0;
	new_entry->keepalive_timer.tv_usec 	= 0;
	new_entry->waiting_for_ping 		= 0;
	new_entry->tun_up 					= 0;
	new_entry->prev_connect_attempt		= NONE;
	new_entry->tunneling_as_mr			= -1;	
	new_entry->cr_deregistered			= 0;
	new_entry->reg_attempt_count		= 0;
	new_entry->udp_tnl_requested		= 0;
	new_entry->regreq_online 			= 0;
	
	/* Open a registration socket (will be binded to CoA random port) */
	new_entry->registration_socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (new_entry->registration_socket < 0) {
		LOG2(LOG_ERR, "get_new_ro_entry - RO Entry's registration socket: %s\n",
		     strerror(errno));
	}
	
	/* Open a socket to send CoTI through and receive CoT from (will be binded to CoA random port) */
	new_entry->coti_socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (new_entry->coti_socket < 0) {
		LOG2(LOG_ERR, "get_new_ro_entry - RO Entry's CoTI socket: %s\n",
		     strerror(errno));
	}
	
	/* Set IP packet infos */
	if (setsockopt(new_entry->registration_socket, SOL_IP, IP_PKTINFO, 
			&one, sizeof(one)) < 0)
	{
    	perror("setsockopt(IP_PKTINFO) - reg socket: ");
	}
	if (setsockopt(new_entry->coti_socket, SOL_IP, IP_PKTINFO, 
			&one, sizeof(one)) < 0)
	{
    	perror("setsockopt(IP_PKTINFO) - CoTI socket: ");
	}
	
	cp_iface_list_into_ro_entry(new_entry);
	
	/* Initially use the first iface in the usable iface list */
	bind_cr_connection_to_device(new_entry, usable_nw_ifaces[0].dev_name);//mn.device_in_use);

	DEBUG(DEBUG_INFO, " - entry's CoA socket fd %d binded to device %s\n",
		new_entry->registration_socket, new_entry->device_in_use); 
	
	new_entry->mr_reg_port = get_socket_local_port(new_entry->registration_socket);
	
	return new_entry;
}

int add_krm_info_into_mr_coas_array(struct ro_entry *cr) {
	int i;
	
	/* Iterate through mr_coas to find the current mr_coa's index */
	for (i = 0; i < cr->mr_coa_count; i++)
	{
		if (cr->mr_coas[i].coa.s_addr == cr->mr_coa.s_addr)
		{
			cr->mr_coas[i].hot_nonce_index	= cr->hot_nonce_index;
			cr->mr_coas[i].cot_nonce_index	= cr->cot_nonce_index;
			cr->mr_coas[i].krm[0] 			= cr->krm[0];
			cr->mr_coas[i].krm[1] 			= cr->krm[1];
			gettimeofday(&cr->mr_coas[i].krm_creation_time, NULL);
			return 0;
		}
	}
	DEBUG(DEBUG_INFO, "add_krm_info_into_mr_coas_array: CoA %s not found in the mr_coas list!\n",
		inet_ntoa(cr->mr_coa));
	return -1;
}

/* delete_ro_entry: Tag RO Entry to be "not in use" by replacing it's 
 * HoA with 0.0.0.0. Remove direct routes towards it, close it's 
 * registration and CoTI sockets and tun_fd. Decrease ro_entry_count.
 * 
 */
int delete_ro_entry(struct ro_entry *cr) {

	int i, del_cr_i;
	
	/* Find out the entry index in the RO Cache */	
	del_cr_i = get_ro_entry_index(cr);
	if (del_cr_i < 0)
		return del_cr_i;
		
	remove_direct_cr_routes(cr);
	
	/* Close the sockets in this RO Entry */
	close(cr->registration_socket);
	close(cr->coti_socket);
	close(cr->tun_fd);
	
	/* Move last entries one index backwards, TODO: optimize using hashtable! */
	for(i = del_cr_i; i <= (ro_entry_count - 2); i++)
	{
		memcpy(&ro_cache[i], &ro_cache[i + 1], sizeof(struct ro_entry));
	}
	
	/* Empty the last element */
	if (ro_entry_count > 0) memset(&ro_cache[ro_entry_count - 1], 0, sizeof(struct ro_entry));

	ro_entry_count--;
	return 0;
}

int clean_ro_bindings() { 
	int i;
	struct ro_entry *cr;
	
	print_trace(mn_traces, "Cleaning up RO Bindings\n");
	
	for (i = 0; i < ro_entry_count; i++) 
	{
		cr = &ro_cache[i];
		remove_direct_cr_routes(cr);
		print_trace(mn_traces, "Deregister to CR %s\n", cr->cr_hoa);
		
		/* Deregister to all CRs */
		send_regreq_to_cr(cr, REG_DISC);
		
		/* Close CR sockets */
		close(cr->registration_socket);
		close(cr->coti_socket);
		close(cr->tun_fd);
	}
	return 0;
}

/* remove_direct_cr_routes: Remove direct routes to CR's HoA and NEMO networks.
 */
int remove_direct_cr_routes(struct ro_entry *cr) {
	int i;
	int system_ret;
	char buf[70];
	
	DEBUG(DEBUG_MESSAGES, "Delete direct routes to CR %s\n", 
		inet_ntoa(cr->cr_hoa));

	for(i = 0; i < cr->cr_nw_count; i++)
	{
		if (dyn_ip_route_exists(&cr->cr_nws[i].prefix, cr->cr_nws[i].prefix_length))
		{
			DEBUG(DEBUG_MESSAGES, "Found direct route to the CR nw %s/%d -> deleting...\n", 
				inet_ntoa(cr->cr_nws[i].prefix), cr->cr_nws[i].prefix_length);
			sprintf(buf, "ip route del %s/%d dev %s", inet_ntoa(cr->cr_nws[i].prefix),
				cr->cr_nws[i].prefix_length, cr->tun_dev_name);
			system_ret = system(buf);
		}
	}
	return 0;
}
/* set_cr_tnl_dst: Set a new dst IP for the tunnel. Also keep 
 * the old tunnel dst alive for DELETED_TUNNEL_UP_TIME 
 */
int set_cr_tnl_dst(struct ro_entry *cr, struct in_addr cr_coa) 
{
	assert(cr != NULL);
	if (cr->tnl_dst.s_addr != cr_coa.s_addr)
	{
		set_timer_expiration(&cr->old_tnl_delete_time, DELETED_TUNNEL_UP_TIME);
		cr->old_tnl_dst.s_addr = cr->tnl_dst.s_addr;
		cr->tnl_dst.s_addr = cr_coa.s_addr; 
	}
	return 0;
}
int iface_is_cr_tnl(char *ifname) {
	int i;
	char *tun_name;
	for (i = 0; i < ro_entry_count; i++) 
	{
		tun_name = ro_cache[i].tun_dev_name;
		if (strcmp(tun_name, ifname) == 0) 
			return 1;
	}
	return 0;
}

int get_ro_entry_index(struct ro_entry *cr) {
	int i;
	assert(cr != NULL);
	for (i = 0; i < ro_entry_count; i++) 
	{
		if (cr->cr_hoa.s_addr == ro_cache[i].cr_hoa.s_addr)
		{
			return i;
		}
		if (i >= ro_entry_count - 1)
		{
			DEBUG(DEBUG_INFO, "get_ro_entry_index: CR HoA %s not found in RO Cache!\n",
				inet_ntoa(cr->cr_hoa));
			return -1;
		}
	}
}
struct ro_entry *get_ro_entry_by_index(int index) {
	if (index < ro_entry_count)
		return &ro_cache[index];
	else
		return NULL;
}
/* get_ro_entry_by_ping_seq: Find out which CR's CoA's tunnel init ping
 * was sent with this sequence number.
 * 
 * Returns: this CR's RO Entry.
 */
struct ro_entry *get_ro_entry_by_ping_seq(unsigned short icmp_seq) {
	int i, p;

	for (i = 0; i < ro_entry_count; i++) {
		for (p = 0; p < KEEPALIVE_ARRAY_LEN; p++) {
			if (ro_cache[i].keepalives[p].timer_on &&
				ro_cache[i].keepalives[p].seq_number == icmp_seq) 
			{
				return &ro_cache[i];
			}
		}
	}
	// No ping to a CR was sent with this sequence number
	return NULL;
}

struct network *get_new_nw_entry_for_ro_entry(struct ro_entry *cr) {
	int new_nw_index;
	
	assert(cr != NULL);

	new_nw_index = cr->cr_nw_count;
	cr->cr_nw_count++;
	
	DEBUG(DEBUG_INFO, "New network in RO Cache Entry: cr->cr_nws[%d]\n",
		new_nw_index);
	
	return &(cr->cr_nws[new_nw_index]);
}
/* get_ro_entry_by_cookie: Get the Route Optimization Cache entry
 * corresponding to this cookie. The parameter can be 
 * either a Home or a Care-of cookie.
 */
struct ro_entry *get_ro_entry_by_cookie(__u64 *cookie) {
	int i;
	if (cookie == NULL) {
		DEBUG(DEBUG_MESSAGES, "get_ro_entry_by_cookie: cookie pointer == NULL!\n");
		return NULL;
	}
	//DEBUG(DEBUG_MESSAGES, "get_ro_entry_by_cookie: cookie %llu\n", ntoh64(*cookie));
	for (i = 0; i < ro_entry_count; i++) {
		//DEBUG(DEBUG_MESSAGES, "ro_cache[%d]: HoTI cookie %llu, CoTI cookie %llu\n", 
		//	i, ntoh64(ro_cache[i].hoti_cookie), ntoh64(ro_cache[i].coti_cookie));
		if (*cookie == ro_cache[i].hoti_cookie ||
			*cookie == ro_cache[i].coti_cookie)
		{
			return &ro_cache[i];
		}
	}
	return NULL;
}
/* get_ro_entry_by_hoa: Get the Route Optimization Cache entry
 * corresponding to the CR Home Address given as parameter.
 */
struct ro_entry *get_ro_entry_by_hoa(struct in_addr hoa) 
{
	int i;
	for (i = 0; i < ro_entry_count; i++) 
	{
		if (hoa.s_addr == ro_cache[i].cr_hoa.s_addr)
		{
			return &ro_cache[i];
		}
	}
	return NULL;	
}

/* get_ro_entry_by_cr_coa: Get the Route Optimization Cache entry
 * corresponding to the CR Care-of Address given as parameter.
 */
struct ro_entry *get_ro_entry_by_cr_coa(struct in_addr cr_coa) {
	int cr_i, coa_i;
	/* Iterate CRs */
	for (cr_i = 0; cr_i < ro_entry_count; cr_i++) 
	{
		/* Iterate current CR's CoAs */
		for (coa_i = 0; coa_i < ro_cache[cr_i].cr_coa_count; coa_i++)
		{
			if (cr_coa.s_addr == ro_cache[cr_i].cr_coas[coa_i].s_addr)
			{
				return &ro_cache[cr_i];
			}
		}
	}
	return NULL;
}

struct ro_entry *get_ro_entry_by_tnl_dst(struct in_addr tnl_dst) {
	int cr_i;
	/* Iterate CRs */
	for (cr_i = 0; cr_i < ro_entry_count; cr_i++) 
	{
		if (ro_cache[cr_i].tnl_dst.s_addr == tnl_dst.s_addr)
		{
			return &ro_cache[cr_i];
		}
	}
	return NULL;
}
int check_cr_old_tnl_alive(struct ro_entry *cr) {
	struct timeval now;
	gettimeofday(&now, NULL);
	if (cmp_timeval(&cr->old_tnl_delete_time, &now) >= 0) 
		return 1;
	else 
		return 0;
}
struct ro_entry *get_ro_entry_by_old_tnl_dst(struct in_addr old_tnl_dst) {
	int cr_i;
	struct timeval now;
	
	gettimeofday(&now, NULL);
	/* Iterate CRs */
	for (cr_i = 0; cr_i < ro_entry_count; cr_i++) 
	{
		/* Check if the address is the old tunnel and that the old tunnel is still alive */
		if (ro_cache[cr_i].old_tnl_dst.s_addr == old_tnl_dst.s_addr &&
			cmp_timeval(&ro_cache[cr_i].old_tnl_delete_time, &now) >= 0)
		{
			return &ro_cache[cr_i];
		}
	}
	return NULL;
}
char *get_cr_tnl_dst_str(struct ro_entry *cr)
{
  	static char buf[50];	/* Static so that the buf can be used by the caller */
	int num_of_chars;
	
	if (cr->tunneling_as_mr) {
		num_of_chars = sprintf(buf, "%s:434", inet_ntoa(cr->tnl_dst));
	} else {
		num_of_chars = sprintf(buf, "%s:%d", inet_ntoa(cr->tnl_dst), ntohs(cr->cr_co_port));
	}
	return buf;
}
/* change_cr_tnl_to_next_coa: Start using the next CR CoA in the 
 * cr_coas list by making it the tnl_dst. 
 * 
 * Returns 
 *  - 0: if ok 
 *  - -1 if error (no more CoAs in the list etc.)
 */
int change_cr_tnl_to_next_coa(struct ro_entry *cr) {
	int i;
	
	/* Check that more than 1 item in cr_coas list */
	if (cr->cr_coa_count <= 1) {
		DEBUG(DEBUG_INFO, "change_cr_tnl_to_next_coa: CR has only %d CoAs!",
			cr->cr_coa_count);
		return -1;
	}
	/* Go through the cr->cr_coas list */
	for (i = 0; i < cr->cr_coa_count; i++)
	{
		/* Find out the tnl_dst index in the list */
		if (cr->cr_coas[i].s_addr == cr->tnl_dst.s_addr) 
		{
			/* If tnl_dst is the last one in the list, take the 
			 * first one in the list. Otherwise take the next one.*/
			if (i >= (cr->cr_coa_count - 1))
			{
				set_cr_tnl_dst(cr, cr->cr_coas[0]);
				//cr->tnl_dst.s_addr = cr->cr_coas[0].s_addr;
			} else {
				set_cr_tnl_dst(cr, cr->cr_coas[i + 1]);
				//cr->tnl_dst.s_addr = cr->cr_coas[i + 1].s_addr;
			}
			return 0;
		}
	}
	DEBUG(DEBUG_INFO, "change_cr_tnl_to_next_coa: tnl_dst not found in the cr_coas list!");
	return -1;
}
/* ro_entry_has_coa:
 * 
 * Returns: 1 if cr_coa is listed in the RO Entry cr,
 *          0 if not.
 */
int ro_entry_has_coa(struct ro_entry *cr, struct in_addr cr_coa) {
	int coa_i;
	/* Iterate current CR's CoAs */
	for (coa_i = 0; coa_i < cr->cr_coa_count; coa_i++)
	{
		if (cr_coa.s_addr == cr->cr_coas[coa_i].s_addr)
		{
			return 1;
		}
	}
	return 0;
}

struct ro_entry *get_ro_entry_by_sent_reg_id(__u32 *reg_id) {
	int i;
	assert(reg_id != NULL);

	DEBUG(DEBUG_MESSAGES, "get_ro_entry_by_reg_id: Look for %08x %08x\n", 
		(__u32) ntohl(reg_id[0]), 
		(__u32) ntohl(reg_id[1]));
		
	for (i = 0; i < ro_entry_count; i++) 
	{
		DEBUG(DEBUG_MESSAGES, "\tro_cache[%d]: HoA %s, sent_req_id %08x %08x, received_req_id %08x %08x\n", 
			i, inet_ntoa(ro_cache[i].cr_hoa),
			(__u32) ntohl(ro_cache[i].sent_reg_id[0]), 
			(__u32) ntohl(ro_cache[i].sent_reg_id[1]),
			(__u32) ntohl(ro_cache[i].received_reg_id[0]), 
			(__u32) ntohl(ro_cache[i].received_reg_id[1]));
			
		/* If reg_id matches the id of the last req sent or last req received */
		if ((reg_id[0] == ro_cache[i].sent_reg_id[0] &&
			 reg_id[1] == ro_cache[i].sent_reg_id[1]))
		{
			return &ro_cache[i];
		}
	}
	return NULL;	
}

struct ro_entry *get_ro_entry_by_reg_id(__u32 *reg_id) {
	int i;
	assert(reg_id != NULL);

	DEBUG(DEBUG_MESSAGES, "get_ro_entry_by_reg_id: Look for %08x %08x\n", 
		(__u32) ntohl(reg_id[0]), 
		(__u32) ntohl(reg_id[1]));
		
	for (i = 0; i < ro_entry_count; i++) 
	{
		DEBUG(DEBUG_MESSAGES, "\tro_cache[%d]: HoA %s, sent_req_id %08x %08x, received_req_id %08x %08x\n", 
			i, inet_ntoa(ro_cache[i].cr_hoa),
			(__u32) ntohl(ro_cache[i].sent_reg_id[0]), 
			(__u32) ntohl(ro_cache[i].sent_reg_id[1]),
			(__u32) ntohl(ro_cache[i].received_reg_id[0]), 
			(__u32) ntohl(ro_cache[i].received_reg_id[1]));
			
		/* If reg_id matches the id of the last req sent or last req received */
		if ((reg_id[0] == ro_cache[i].sent_reg_id[0] &&
			 reg_id[1] == ro_cache[i].sent_reg_id[1]) ||
			(reg_id[0] == ro_cache[i].received_reg_id[0] &&
			 reg_id[1] == ro_cache[i].received_reg_id[1]))
		{
			return &ro_cache[i];
		}
	}
	return NULL;	
}

int clear_cr_network_list(struct ro_entry *cr) {
	cr->cr_nw_count = 0;
	return 0;
}

/* ro_entry_has_nw: Check if RO Entry has the nw listed.
 * 
 * Returns: 
 *  - 0 if nw is not listed for the CR
 *  - 1 if nw is listed and already authorized
 *  - 2 if nw is listed but not yet authorized
 */
int ro_entry_has_nw(struct ro_entry *cr, struct in_addr nw) {
	int i;
	
	for(i = 0; i < cr->cr_nw_count; i++)
	{
		if (cr->cr_nws[i].prefix.s_addr == nw.s_addr)
		{
			if (cr->cr_nws[i].authorized)
				return 1;
			else
				return 2;
		}
	}
	return 0;
}

/* add_nw_for_ro_entry: Add a nw in the end of the cr_nws array unless 
 * it exists there already. 
 *
 * Returns: 
 *  - 0 if nw added as new
 *  - 1 if nw found from the list as authorized
 *  - 2 if nw found from the list as unauthorized
 */
int add_unconfirmed_nw_into_ro_entry(struct ro_entry *cr, struct in_addr *nw, int plen) 
{
	int nw_check_ret;
	
	assert(cr != NULL);
	
	/* If nw is in the list already -> return without modifying */
	nw_check_ret = ro_entry_has_nw(cr, *nw);
	if (nw_check_ret == 0)
	{
		cr->cr_nws[cr->cr_nw_count].prefix.s_addr 	= nw->s_addr;
		cr->cr_nws[cr->cr_nw_count].prefix_length 	= plen;
		if (plen >= 0 && plen <= 32)
			cr->cr_nws[cr->cr_nw_count].plen_ok		= 1;
		else
			cr->cr_nws[cr->cr_nw_count].plen_ok		= 0;
		//cr->cr_nws[cr->cr_nw_count].forwarding_ok	= forwarding_ok;
		cr->cr_nws[cr->cr_nw_count].authorized		= 0;
		cr->cr_nw_count++;
	}
	return nw_check_ret;
}

/* add_nw_item_in_ro_entry: Add or update a nw item in the RO Entry. This can be 
 * used if a nw is advertised by HA. */
int add_confirmed_nw_item_into_ro_entry(struct ro_entry *cr, struct network *nw) {
	int i;
	
	for(i = 0; i < cr->cr_nw_count; i++)
	{
		/* Nw already in the entry -> replace */
		if (cr->cr_nws[i].prefix.s_addr == nw->prefix.s_addr)
		{
			memcpy(&cr->cr_nws[i], nw, sizeof(struct network));
			DEBUG(DEBUG_MESSAGES, "Update CR %s nw[%d] ", inet_ntoa(cr->cr_hoa), i);
			DEBUG(DEBUG_MESSAGES, "%s/%d\n", inet_ntoa(cr->cr_nws[i].prefix), 
				cr->cr_nws[i].prefix_length);
			return 1;
		}
	}
	/* Nw not found from the entry -> add */
	memcpy(&cr->cr_nws[cr->cr_nw_count], nw, sizeof(struct network));
	DEBUG(DEBUG_MESSAGES, "Add nw for CR %s: nw[%d] ", inet_ntoa(cr->cr_hoa), cr->cr_nw_count);
	DEBUG(DEBUG_MESSAGES, "%s/%d\n", inet_ntoa(cr->cr_nws[cr->cr_nw_count].prefix), 
		cr->cr_nws[cr->cr_nw_count].prefix_length);
	
	cr->cr_nw_count++;
	
	return 0;
}
/* add_cr_coa_into_ro_entry:  Add CR CoA into RO Entry if it's not there already.
 * 
 * @cr: pointer to the RO Entry.
 * @cr_coa: CR CoA to add. 
 * 
 */
int add_cr_coa_into_ro_entry(struct ro_entry *cr, struct in_addr cr_coa)
{
	if (ro_entry_has_coa(cr, cr_coa)) {
		return -1;
	} else {
		assert(cr != NULL);
		cr->cr_coas[cr->cr_coa_count].s_addr = cr_coa.s_addr; 
		
		DEBUG(DEBUG_MESSAGES, "add_cr_coa_into_ro_entry: Add CoA[%d] %s ", 
			cr->cr_coa_count, inet_ntoa(cr->cr_coas[cr->cr_coa_count]));
		DEBUG(DEBUG_MESSAGES, "for CR %s\n", 
			inet_ntoa(cr->cr_hoa));
		
		cr->cr_coa_count++;
		
		/* If this is the first CoA, make it the tunnel dst */
		if (cr->cr_coa_count <= 1)
			set_cr_tnl_dst(cr, cr_coa);
		
		return 0;
	}
}

/* TODO: Make the MR RO Entry states more clear */

/* go_to_resend_hoti_coti_state: Put the RO Cache Entry to a state where new 
 * HoTI and CoTI messages are sent and KRm is recalculated.
 */
int go_to_resend_hoti_coti_state(struct ro_entry *cr)
{
	cr->hot_received 			= 0;
	cr->cot_received 			= 0;
	cr->send_init_msgs 			= 1;
	cr->different_coti_iface 	= 0;
	return 0;
}
int go_to_resend_hoti_coti_through_new_iface_state(struct ro_entry *cr)
{
	cr->hot_received 			= 0;
	cr->cot_received 			= 0;
	cr->send_init_msgs 			= 1;
	cr->different_coti_iface 	= 1;
	return 0;
}
int check_if_plen_ok(int plen) 
{
	if (plen >= 0 && plen <= 32) 
		return 1;
	else
		return 0; 
}
/* Get next ICMP sequence number free to use. 
 * 
 */
unsigned short get_new_icmp_seq() {
	unsigned short return_seq;
	
	if (mn.icmp_seq_num == 0xffff)
		mn.icmp_seq_num = 1;
	
	return_seq = mn.icmp_seq_num; 
	mn.icmp_seq_num++;
	
	return return_seq;
}
/*
 * Determine the timeout value in seconds for a solicitation request.
 * Parameters:
 *  counter - pointer to variable holding information,
 *            how many requests have been sent.
 * Returns: timeout in seconds or zero when it's time to go passive mode.
 */
static int get_solicitation_interval(int *counter)
{
	static int intervals[] = SOLICITATION_INTERVALS;
	int t;

	ASSERT(counter != NULL);
	ASSERT(sizeof(intervals) > 0);

	if (*counter < 0) *counter = 0;

	if (*counter >= sizeof(intervals) / sizeof(int))
		t = 0;
	else t = intervals[*counter];

	(*counter)++;

	return t;

}

/*
 * Due to the unsuccessful registration to the FA, degrade its priority.
 */
int degrade_current_fa_priority(void)
{
	if (mn.current_adv == NULL)
		return 1;

	if (mn.current_adv->prio_degrade_percent == 0) {
		mn.current_adv->prio_degrade_percent = 
			DEFAULT_PRIO_DEGRADE_INITIAL;
	} else {
		mn.current_adv->prio_degrade_percent *= 
			DEFAULT_PRIO_DEGRADE_FACTOR;
	}

	if (mn.current_adv->prio_degrade_percent > 100) {
		mn.current_adv->prio_degrade_percent = 100;
	}

	DEBUG(DEBUG_INFO, "degrade_current_fa_priority: %d\n",
	      mn.current_adv->prio_degrade_percent);
	
	return 0;
}


/*
 * Set reregistration time. Binary exponential backoff is used on
 * reregistrations. Maximum reregistration time is limited with
 * NORMAL_REREGISTRATION_TIME. At least half of the lifetime is
 * allowed to pass before registration is initiated.
 */
static void set_reregistration_time(int total_lifetime)
{
	int rt;   /* remaining lifetime */
	struct timeval expire_time, now;
	struct fa_spi_entry *fa_spi;

	expire_time = timers[TIMER_LIFETIME];

	gettimeofday(&now, NULL);
	DEBUG(DEBUG_TIMERS, "set_reregistration_time(%i) - now=%li.%06li\n",
	      total_lifetime, now.tv_sec, now.tv_usec);
	DEBUG(DEBUG_TIMERS, "\tTIMER_LIFETIME: %li.%06li\n",
	      timers[TIMER_LIFETIME].tv_sec, timers[TIMER_LIFETIME].tv_usec);

	if (config.mn_ha_key_timestamp != 0 &&
	    config.mn_ha_key_lifetime != 0 &&
	    expire_time.tv_sec > config.mn_ha_key_timestamp +
	    config.mn_ha_key_lifetime) {
		expire_time.tv_sec = config.mn_ha_key_timestamp +
			config.mn_ha_key_lifetime;
		expire_time.tv_usec = 0;
		mn.aaa_rekey = 1;

		DEBUG(DEBUG_TIMERS, "\tMN-HA key lifetime overriding at "
		      "%li\n", expire_time.tv_sec);
	}

	fa_spi = get_fa_spi(0, mn.fa_addr);
	if (fa_spi != NULL && fa_spi->created > 0 &&
	    expire_time.tv_sec > fa_spi->created + fa_spi->lifetime) {
		expire_time.tv_sec = fa_spi->created + fa_spi->lifetime;
		expire_time.tv_usec = 0;
		mn.aaa_rekey = 1;

		DEBUG(DEBUG_TIMERS, "\tMN-FA key lifetime overriding at "
		      "%li\n", expire_time.tv_sec);
	}

	rt = sec_passed(&expire_time);

	/* if total lifetime is set ( > 0), set initial reregistration
	 *  time, else use exponential backoff
	 */
	if (total_lifetime > 0) {
		/* bound to end of the lifetime */
		if (rt > 2*NORMAL_REREGISTRATION_TIME) {
			timers[TIMER_REREG] = expire_time;
			timers[TIMER_REREG].tv_sec -=
				mn.aaa_rekey ? MN_AAA_REG_TIME :
				NORMAL_REREGISTRATION_TIME;
		} else {
			struct timeval rereg;

			now.tv_sec++;

			rereg = expire_time;
			rereg.tv_sec -= mn.aaa_rekey ? MN_AAA_REG_TIME :
				total_lifetime / 2;

			if (cmp_timeval(&rereg, &now) > 0)
				timers[TIMER_REREG] = rereg;
			else
				timers[TIMER_REREG] = now;
		}
		if (mn.current_adv == NULL) {
			LOG2(LOG_ALERT, "set_registration_time - current_adv "
			     "== NULL\n");
			mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;
		} else {
			DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time = %i\n",
			      inet_ntoa(mn.current_adv->addr),
			      MIN_REGISTRATION_TIME);
			mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
		}
	} else {
		if (mn.current_adv == NULL) {
			LOG2(LOG_ALERT, "set_registration_time - current_adv "
			     "== NULL\n");
			if (mn.HA_reg_retry_time == 0)
				mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;
			else
				mn.HA_reg_retry_time *= 2;
			timers[TIMER_REREG].tv_sec += mn.HA_reg_retry_time;
		} else {
#ifndef NO_PRIO_DEGRADE
			/* degrade FA priority */
			degrade_current_fa_priority();
#endif /* NO_PRIO_DEGRADE */

			DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: %i => ",
			      inet_ntoa(mn.current_adv->addr),
			      mn.current_adv->reg_retry_time);
			if (mn.current_adv->reg_retry_time == 0)
				mn.current_adv->reg_retry_time =
					MIN_REGISTRATION_TIME;
			else
				mn.current_adv->reg_retry_time *= 2;
			DEBUG(DEBUG_TIMERS, "%i\n",
			      mn.current_adv->reg_retry_time);
			timers[TIMER_REREG].tv_sec +=
				mn.current_adv->reg_retry_time;
		}
	}
}


/* Select next timer to timeout and place timeout to *tv.
 * Returns non-zero if timeout is valid otherwise 0. 
 */
static int get_next_timeout(struct timeval *tv)
{
	struct timeval *t, now, difference;
	int i, cr_i, k;
	struct ro_entry *cr;
	
	gettimeofday(&now, NULL);
	
	/* t = general_timer */
	t = &timers[TIMER_GEN];

	/* Print some current timeout intervals */
	DEBUG(DEBUG_TIMERS, "get_next_timeout: \n");	

	/*------------------------------------------------------*
	 | Place the next timeout in the variable t.			|
	 *------------------------------------------------------*/
	/* If the timer is set and smaller than the current t
	 *   -> replace t with the timer */
	
	/* HA keepalive timer */
	if ( timerisset(&timers[TIMER_KEEPALIVE]) && 
		(!timerisset(t) || cmp_timeval(&timers[TIMER_KEEPALIVE], t) < 0) )
	{
		t = &timers[TIMER_KEEPALIVE];
		
		timersub(&timers[TIMER_KEEPALIVE], &now, &difference);
		DEBUG(DEBUG_TIMERS, " - HA keepalive timer to expire next (in %li ms).\n",
				difference.tv_usec / 1000);
	}
	/* HA Reply wait timer */
	if ( timerisset(&mn.reg_retry_time) && 
		(!timerisset(t) || cmp_timeval(&mn.reg_retry_time, t) < 0) )
	{
		t = &mn.reg_retry_time;
		
		timersub(&mn.reg_retry_time, &now, &difference);
		DEBUG(DEBUG_TIMERS, " - HA Reply wait timer to expire next (in %li ms).\n",
				difference.tv_usec / 1000);
	}
	
	/* HA ping failure timers */
	//DEBUG(DEBUG_TIMERS, "RTTs: [%d %d %d %d %d], average %d\n", mn.RTTs[0], mn.RTTs[1], 
	//	mn.RTTs[2], mn.RTTs[3], mn.RTTs[4], get_average_RTT_to_ha());
	for (i = 0; i < KEEPALIVE_ARRAY_LEN; i++) {
		if (mn.ha_keepalives[i].timer_on) 
		{
			if (!timerisset(t) || 
				cmp_timeval(&mn.ha_keepalives[i].failure_timer, t) < 0) 
			{
				t = &mn.ha_keepalives[i].failure_timer;
				
				timersub(&mn.ha_keepalives[i].failure_timer, &now, &difference);
				DEBUG(DEBUG_TIMERS, " - Ping (seq %d) failure timer to expire next (in %li ms).\n",
					mn.ha_keepalives[i].seq_number, difference.tv_usec / 1000);
			}
		}
	}
	
	/* CRs' timers */
	for (cr_i = 0; cr_i < ro_entry_count; cr_i++) 
	{
		cr = (struct ro_entry *) get_ro_entry_by_index(cr_i);
		
		/* CR keepalive timer */
		if (timerisset(&cr->keepalive_timer) && 
			(!timerisset(t) || cmp_timeval(&cr->keepalive_timer, t) < 0) )
		{
			t = &cr->keepalive_timer;
			
			timersub(&cr->keepalive_timer, &now, &difference);
			DEBUG(DEBUG_TIMERS, " - CR %s keepalive timer is to expire next (in %li ms).\n", 
				inet_ntoa(cr->cr_hoa), difference.tv_usec / 1000);
		}
		/* CR Reply Wait timer */
		if (timerisset(&cr->reply_wait_timer) && 
			(!timerisset(t) || cmp_timeval(&cr->reply_wait_timer, t) < 0) )
		{
			t = &cr->reply_wait_timer;
			
			timersub(&cr->reply_wait_timer, &now, &difference);
			DEBUG(DEBUG_TIMERS, " - CR %s Reply waiting timer is to expire next (in %li ms).\n", 
				inet_ntoa(cr->cr_hoa), difference.tv_usec / 1000);
		}
		/* CR ping failure timers */
		for (k = 0; k < KEEPALIVE_ARRAY_LEN; k++) 
		{
			if (cr->keepalives[k].timer_on) 
			{
				if (!timerisset(t) || 
					cmp_timeval(&cr->keepalives[k].failure_timer, t) < 0) 
				{
					t = &cr->keepalives[k].failure_timer;
					
					timersub(&cr->keepalives[k].failure_timer, &now, &difference);
					DEBUG(DEBUG_TIMERS, " - Ping (seq %d) to CR (%s) failure timer to expire next (in %li ms).\n",
						cr->keepalives[k].seq_number, inet_ntoa(cr->cr_hoa), difference.tv_usec / 1000);
				}
			}
		}
	}
	
	/* Reregistration timer */
	if (timerisset(&timers[TIMER_REREG]) 	&& 
		(!timerisset(t) || 
		cmp_timeval(&timers[TIMER_REREG], t) < 0))
	{
		t = &timers[TIMER_REREG];
		DEBUG(DEBUG_TIMERS, " - Rereg timer to expire next.\n");
	}
	
	if (config.wlan_ap_poll_interval > -1 		&&
	    timerisset(&timers[TIMER_WLAN_AP_POLL]) &&
	    (!timerisset(t) ||
	     cmp_timeval(&timers[TIMER_WLAN_AP_POLL], t) < 0))
	{
		DEBUG(DEBUG_TIMERS, " - WLAN poll timer to expire next.\n");
		t = &timers[TIMER_WLAN_AP_POLL];
	}
	
	if (config.solicitation_interval > -1 &&
	    timerisset(&timers[TIMER_SOLICITATION]) &&
	    (!timerisset(t) ||
	     cmp_timeval(&timers[TIMER_SOLICITATION], t) < 0))
	{
		DEBUG(DEBUG_TIMERS, " - Agent solicitation timer to expire next.\n");
		t = &timers[TIMER_SOLICITATION];
	}

	if (!timerisset(t)) return 0;

	/* check if the current agentadv expires before the TIMER_GEN */
	if (mn.current_adv && mn.expire_check &&
	    cmp_timeval(&mn.current_adv->expire, t) < 0) 
	{
		DEBUG(DEBUG_TIMERS, " - Agent adv. timer to expire next.\n");
		mn.expire_check = 0;
		t = &mn.current_adv->expire;
	}

	if (now.tv_usec > t->tv_usec) {
		tv->tv_sec 	= t->tv_sec - now.tv_sec - 1;
		tv->tv_usec = 1000000 	+ t->tv_usec - now.tv_usec;
	} else {
		tv->tv_sec 	= t->tv_sec 	- now.tv_sec;
		tv->tv_usec = t->tv_usec 	- now.tv_usec;
	}

	if (tv->tv_sec < 0) timerclear(tv);

	return 1;
}


/* State change functions
 * Same functions are used for entering the for the state first time and
 * retrying after timeout. If state is already set, timeout processing is done.
 */

/* Just listen to agent advertisements and switch to request_tunnel after one
 * is heard. */
void passive_find(void)
{
        mn.state = MN_PASSIVE_FIND;
        DEBUG(DEBUG_STATES, "State: Passive find\n");
}


static void send_solicitations(struct mn_data *mn)
{
	int i;
	struct timeval now;

	gettimeofday(&now, NULL);
	for (i = 0; i < MAX_INTERFACES; i++) {
		if (mn->iface[i].s > -1) {
			if ((now.tv_sec < mn->iface[i].last_solicitation.tv_sec
			     || !timerisset(&mn->iface[i].last_solicitation) ||
			     usec_passed(&mn->iface[i].last_solicitation, &now)
			     > MIN_SOLICITATION_DELAY)) {
				send_agent_solicitation(mn->iface[i].s);
				mn->iface[i].last_solicitation = now;
			} else {
				DEBUG(DEBUG_TIMERS, "Too frequent agent "
				      "solicitations (dev=%s) - skipping\n",
				      mn->iface[i].device);
			}
		}
	}
}

/**
 * find_agent:
 * @entry - STATE_INIT to move to this state
 *          STATE_TIMEOUT to indicate timeout in this state,
 *          ie. we haven't received an agent adv to our last sent
 *          solicitation
 *
 *  Enter find agent state
 */
void find_agent(int entry)
{
	int t, chg;
	static int retry_time = 0; /* time in seconds between retries of agent
							    * solicitations */

	DEBUG(DEBUG_INFO, "find_agent(%i)\n", entry);

	mn.tunnel_up = 0;

	if (mn.device_count == 0) {
		passive_find();
		return;
	}

	if (mn.tunnel_mode == API_TUNNEL_FULL_HA) {
		DEBUG(DEBUG_INFO,
		      "find_agent: trying to register directly to HA\n");
		if (config.ha_ip_addr.s_addr == 0 &&
		    config.home_net_addr_plen > -1) {
			/* dynamic HA address resolution */
			mn.fa_addr.s_addr = config.home_net_subnet_bc.s_addr;
		} else
			mn.fa_addr.s_addr = config.ha_ip_addr.s_addr;
		mn.co_addr.s_addr = mn.local_addr.s_addr;
		if (mn.current_adv != NULL) {
			mn.current_adv->in_use = 0;
			mn.current_adv = NULL;
		}
		request_tunnel(STATE_INIT, 0, 0);
		return;
	}

	/* try to find the best FA from the previously received agentadvs;
	 * if a valid entry is found, use it; otherwise send agent
	 * solicitation */
	chg = get_fa(NULL);
	if (chg != FA_GET_NO) {
		DEBUG(DEBUG_INFO, "find_agent: found FA from stored agentadv "
		      "data - FA=%s, HFA=", inet_ntoa(mn.fa_addr));
		DEBUG(DEBUG_INFO, "%s\n", inet_ntoa(mn.co_addr));
		request_tunnel(STATE_INIT, 0, 0);
		return;
	}

	if (entry == STATE_INIT) {
                mn.state = MN_FIND_AGENT;
                retry_time = 0;
		mn.req_lifetime = config.mn_default_tunnel_lifetime;
	}

	t = get_solicitation_interval(&retry_time);
	if (t == 0) {
		passive_find();
		return;
	}

	/* send agent solicitation to all the active interfaces */
	send_solicitations(&mn);

	DEBUG(DEBUG_TIMERS, "find_agent - TIMER_GEN: set now+%i sec\n", t);
	gettimeofday(&timers[TIMER_GEN], NULL);
	timers[TIMER_GEN].tv_sec += t;
	add_usecs(&timers[TIMER_GEN],
		  get_rand32() % MAX_RANDOM_SOLICITATION_DELAY);

	DEBUG(DEBUG_STATES, "State: Find agent\n");
}


/**
 * request_tunnel:
 * @entry: STATE_INIT to move to this state,
 *         STATE_TIMEOUT to indicate timeout in this state,
 *         i.e. no reply for last request within retry time
 *
 * Enter request tunnel state
 */
void request_tunnel(int entry, int forced, int check_timer)
{
	int retry_time;

	DEBUG(DEBUG_INFO, "request_tunnel(%i, %i, %i)\n", entry, forced,
	      check_timer);

	if (mn.current_adv == NULL) {
		retry_time = mn.HA_reg_retry_time;
		DEBUG(DEBUG_INFO, "No current agent advertisement " 
		      "when requesting tunnel (direct tunnel to HA?)\n");
	} else
		retry_time = mn.current_adv->reg_retry_time;

	if (check_timer && retry_time > 1) {
		struct timeval now;
		gettimeofday(&now, NULL);
		if (timerisset(&timers[TIMER_GEN]) &&
		    cmp_timeval(&now, &timers[TIMER_GEN]) < 0) {
			DEBUG(DEBUG_TIMERS, "request_tunnel: TIMER_GEN not "
			      "yet reached\n");
			return;
		}
	}

	if (config.enable_fa_decapsulation && mn.current_adv != NULL)
		add_fa_host_route(mn.current_adv->ifname, mn.agentadv,
				  mn.current_adv->ifindex, mn.fa_addr);

        if (entry == STATE_INIT) {
                mn.state = MN_REQUEST_TUNNEL;
		if (!forced)
			mn.tunnel_up = 0;

		/* if last registration was sent less than a second ago,
		   wait a second before sending the request */
		if (usec_passed(&mn.last_reg_send_time, NULL) <
		    MIN_REGISTRATION_DELAY) {
			DEBUG(DEBUG_INFO, "Too frequent registration request "
			      "- delaying\n");
			DEBUG(DEBUG_TIMERS, "request_tunnel: TIMER_GEN: set "
			      "now+1 sec (too freq.)\n");
			gettimeofday(&timers[TIMER_GEN], NULL);
			timers[TIMER_GEN].tv_sec++;
			return;
		}
        } else if (retry_time > MAX_REGISTRATION_TIME) {
		/* STATE_TIMEOUT */
		if (config.use_hadisc) {
			DEBUG(DEBUG_INFO, "HA does not reply - try to discover"
			      " another HA\n");
			mn.info_str = "restarting HA discovery";
			config.ha_ip_addr.s_addr = 0;
			if (mn.current_adv != NULL)
				mn.current_adv->reg_retry_time =
					MIN_REGISTRATION_TIME;
			else
				mn.HA_reg_retry_time =
					MIN_REGISTRATION_TIME;
		} else {
			/* give up, switch to find agent */
			if (forced) {
				DEBUG(DEBUG_INFO,
				      "Could not confirm tunnel.\n");
				reply_waiting_api(API_FAILED, NULL, 0);
			} else {
				DEBUG(DEBUG_INFO, "Could not register.\n");
			}
			find_agent(STATE_INIT);
			return;
		}
        } else if (mn.current_adv != NULL) {
		/* STATE_TIMEOUT */
#ifndef NO_PRIO_DEGRADE
		/* degrade the FA priority */
		degrade_current_fa_priority();
#endif /* NO_PRIO_DEGRADE */
		DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: %i => ",
		      inet_ntoa(mn.current_adv->addr),
		      mn.current_adv->reg_retry_time);
		if (mn.current_adv->reg_retry_time == 0)
			mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
		else
			mn.current_adv->reg_retry_time *= 2;
		DEBUG(DEBUG_TIMERS, "%i\n", mn.current_adv->reg_retry_time);
	} else {
		if (mn.HA_reg_retry_time == 0)
			mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;
		else
			mn.HA_reg_retry_time *= 2;
	}

	if (send_registration(forced ? REG_REREG : REG_CONNECT) < 0) {
		DEBUG(DEBUG_INFO, "Registration failed for this FA - trying to"
		      " find a new one\n");
		mn.warn_str = "Registration failed for this FA, trying to find"
			" a new one";
		find_agent(STATE_INIT);
		return;
	}

        gettimeofday(&timers[TIMER_GEN], NULL);
	if (mn.current_adv != NULL) {
		if (mn.current_adv->reg_retry_time < MIN_REGISTRATION_TIME)
			mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
		retry_time = mn.current_adv->reg_retry_time;
	} else {
		if (mn.HA_reg_retry_time < MIN_REGISTRATION_TIME)
			mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;
		retry_time = mn.HA_reg_retry_time;
	}
	DEBUG(DEBUG_TIMERS, "request_tunnel - TIMER_GEN: set now+%i sec\n",
	      retry_time);
	timers[TIMER_GEN].tv_sec += retry_time;

        DEBUG(DEBUG_STATES, "State: Request tunnel%s\n",
	      forced ? " (forced)" : "");
}


/* Enter connected state.
 * Paramters:
 *   type:
 *     CON_HA   : connection was approved by HA
 *     CON_FA   : connection was approved by FA
 *   lifetime: lifetime received in reply
 */
void connected(int type, __u16 lifetime)
{
	DEBUG(DEBUG_INFO, "connected(%i, %i)\n", type, lifetime);
	ASSERT(lifetime > 0);

	if (mn.current_adv != NULL) {
		if (mn.current_adv->addr.s_addr != mn.fa_addr.s_addr) {
			DEBUG(DEBUG_INFO, "fa_addr=%s != ",
			      inet_ntoa(mn.fa_addr));
			DEBUG(DEBUG_INFO, "current_adv->addr=%s\n",
			      inet_ntoa(mn.current_adv->addr));
		}
		mn.current_adv->in_use = 1;
		/* This FA works, so don't degrade priority: */
		mn.current_adv->prio_degrade_percent = 0; 
	} else {
		LOG2(LOG_WARNING, "connected - current_adv == NULL\n");
		mn.warn_str = "connected - current_adv == NULL";
	}

	/* HA MUST NOT use larger lifetime and the FAs cannot change the
	 * requested lifetime - force the lifetime to be at maximum the
	 * configured time */
	if (lifetime > config.mn_default_tunnel_lifetime) {
		LOG2(LOG_WARNING, "Lifetime in the reply (%i sec) larger than "
		     "expected (%i sec) - lowering it\n", lifetime,
		     config.mn_default_tunnel_lifetime);
		mn.warn_str = "lifetime in the reply larger than expected";
		lifetime = config.mn_default_tunnel_lifetime;
	}

	if (config.enable_fa_decapsulation &&
	    mn.fa_addr.s_addr != mn.tunnel_addr.s_addr &&
	    mn.tunnel_addr.s_addr != 0 &&
	    mn.current_adv != NULL)
		remove_fa_host_routes(0);

    /* if tunnel endpoint has changed, or tunnel is down, create tunnel to
	 * FA */
	if (mn.fa_addr.s_addr != mn.tunnel_addr.s_addr || !real_tunnel_up) 
	{
		if (real_tunnel_up) {
			DEBUG(DEBUG_STATES, "Restart tunneling\n");
			restart_tunneling();
		} else {
			DEBUG(DEBUG_STATES, "Start tunneling\n");
			start_tunneling();
		}
	} else {
		/* If tunnel is up, start sending keepalives */
		send_ping_to_HA();
	}
	/* set lifetime timer */
	timers[TIMER_LIFETIME].tv_sec = timers[TIMER_REQUEST].tv_sec +
		lifetime;
	timers[TIMER_LIFETIME].tv_usec = timers[TIMER_REQUEST].tv_usec;

	  /* If connection was approved by HA,
	   * reset timer for reregistering the tunnel */
	if (type == CON_HA) {
		set_reregistration_time(lifetime);
	}
	if (mn.state != MN_CONNECTED && type == CON_HA) {
		LOG2(LOG_INFO, "Connection established.\n");
		mn.info_str = "connection established";
	}
	/* jpaaso - caused problems */
	/*
	DEBUG(DEBUG_TIMERS, "setting TIMER_GEN = TIMER_REREG %li.%06li (in %li"
	      " sec)\n",
	      timers[TIMER_REREG].tv_sec, timers[TIMER_REREG].tv_usec,
	      timers[TIMER_REREG].tv_sec - time(NULL));
	timers[TIMER_GEN] = timers[TIMER_REREG];
	*/
	timers[TIMER_GEN] = timers[TIMER_KEEPALIVE];
	DEBUG(DEBUG_TIMERS, "setting TIMER_GEN = TIMER_KEEPALIVE %li.%06li (in %li"
	      " sec)\n",
	      timers[TIMER_KEEPALIVE].tv_sec, timers[TIMER_KEEPALIVE].tv_usec,
	      timers[TIMER_KEEPALIVE].tv_sec - time(NULL));
	
	mn.state = MN_CONNECTED;
	mn.tunnel_up = 1;
	DEBUG(DEBUG_STATES, "State: Connected\n");
}


/* Send registration request and set timer */
static void reregister(void)
{
	struct timeval now;

	DEBUG(DEBUG_INFO, "Reregistering..\n");

	/* if tunnel has expired, switch to find agent */
	gettimeofday(&now, NULL);
	if (cmp_timeval(&timers[TIMER_LIFETIME], &now) <= 0) {
		DEBUG(DEBUG_INFO, "Binding expired - trying to find another "
		      "FA and re-establish binding\n");
		mn_remove_dynamic_home_addr();
		
		/* make sure that closest FA won't take next reg_req
		 * as locupd. */
		if (mn.session_key != NULL) free(mn.session_key);
		mn.session_key = NULL;
		mn.session_key_len = 0;
                find_agent(STATE_INIT);
		return;
	}

	if (send_registration(REG_REREG) < 0) {
		mn.warn_str = "Registration failed, trying to find another FA";
		DEBUG(DEBUG_INFO,
		      "Registration failed - trying to find another FA\n");
		if (mn.session_key != NULL) free(mn.session_key);
		mn.session_key = NULL;
		mn.session_key_len = 0;
                find_agent(STATE_INIT);
		return;
	}

	set_reregistration_time(0);

	if (cmp_timeval(&timers[TIMER_REREG], &timers[TIMER_LIFETIME]) < 0) {
		DEBUG(DEBUG_TIMERS, "setting TIMER_GEN = TIMER_REREG %li.%06li"
		      " (in %li sec)\n", timers[TIMER_REREG].tv_sec,
		      timers[TIMER_REREG].tv_usec,
		      timers[TIMER_REREG].tv_sec - time(NULL));
		timers[TIMER_GEN] = timers[TIMER_REREG];
	} else {
		DEBUG(DEBUG_TIMERS, "setting TIMER_GEN = TIMER_LIFETIME "
		      "%li.%06li (in %li sec)\n",
		      timers[TIMER_LIFETIME].tv_sec,
		      timers[TIMER_LIFETIME].tv_usec,
		      timers[TIMER_REREG].tv_sec - time(NULL));
		timers[TIMER_GEN] = timers[TIMER_LIFETIME];
	}

}


/* Stop tunneling and send deregistering message to the home agent.
 * We'll advance to next state when reply arrives. If the message is lost, it
 * will be resent after an interval.
 * Parameters:
 *  entry - STATE_INIT to move to this state
 *          STATE_TIMEOUT to indicate timeout in this state
 */
void close_for_home(int entry)
{
	int retry_time;

	DEBUG(DEBUG_INFO, "close_for_home(%i)\n", entry);

	if (mn.current_adv == NULL) {
		LOG2(LOG_ALERT, "close_for_home - current_adv == NULL\n");
	}
	if (entry == STATE_INIT) {
		mn.state = MN_CLOSE_FOR_HOME;
		if (real_tunnel_up)
			stop_tunneling();
		else if (config.shared_secret_len < 0 ||
			 config.mn_home_ip_addr.s_addr == 0) {
			DEBUG(DEBUG_INFO,
			      "close_for_home - not registered; going directly"
			      " to home\n");
			at_home();
			return;
		} else {
			DEBUG(DEBUG_INFO,
			      "close_for_home - not registered, but trying to "
			      "deregister possible old bindings at HA\n");
			if (mn.co_addr.s_addr == 0)
				mn.co_addr.s_addr = mn.local_addr.s_addr;
		}
		if (mn.current_adv != NULL) {
			DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: "
			      "%i => %i (close_for_home)\n",
			      inet_ntoa(mn.current_adv->addr),
			      mn.current_adv->reg_retry_time,
			      MIN_REGISTRATION_TIME);
			mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
		} else
			mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;

		/* drop the previous session key so that locupd to FA with the
		 * old binding goes to the HA */
		if (mn.session_key != NULL)
			free(mn.session_key);
		mn.session_key = NULL;
		mn.session_key_len = 0;
	} else if (mn.current_adv == NULL ||
		   mn.current_adv->reg_retry_time >= MAX_REGISTRATION_TIME) {
		DEBUG(DEBUG_MESSAGES,
		      "Could not get reply from HA - assuming deregistration "
		      "worked\n");
		at_home();
		return;
	} else if (mn.current_adv != NULL) {
		DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: %i => ",
		      inet_ntoa(mn.current_adv->addr),
		      mn.current_adv->reg_retry_time);
		if (mn.current_adv->reg_retry_time == 0)
			mn.current_adv->reg_retry_time = 1;
		else
			mn.current_adv->reg_retry_time *= 2;
		DEBUG(DEBUG_TIMERS, "%i\n", mn.current_adv->reg_retry_time);
	}

	send_gratuitous_arp(config.ha_ip_addr, config.mn_home_ip_addr);

	/* send deregistration message to HA */
	mn.fa_addr = config.ha_ip_addr;
	send_registration(REG_DISC);

	/* at least full retry time should be waited */
	gettimeofday(&timers[TIMER_GEN], NULL);
	retry_time = mn.current_adv ? mn.current_adv->reg_retry_time :
		mn.HA_reg_retry_time;
	DEBUG(DEBUG_TIMERS, "close_for_home - TIMER_GEN: set now+%i sec\n",
	      retry_time);
	timers[TIMER_GEN].tv_sec += retry_time;

        DEBUG(DEBUG_STATES, "State: Close for home\n");
}


/* We have successfully deregistered from HA and wait for IP address change
 * signal to start tunneling again. */
void at_home(void)
{
	DEBUG(DEBUG_INFO, "at_home()\n");
	send_gratuitous_arp(config.ha_ip_addr, config.mn_home_ip_addr);
	mn.state = MN_AT_HOME;
    DEBUG(DEBUG_STATES, "State: At home\n");
	DEBUG(DEBUG_TIMERS, "Clearing TIMER_GEN\n");
	timerclear(&timers[TIMER_GEN]);
	mn_remove_dynamic_home_addr();
}


/* We are disconnected from tunneling system and nothing is done until
 * connection is asked. */
void disconnect(void)
{
	DEBUG(DEBUG_INFO, "disconnect()\n");

	mn.cur_route_info.known = 0;

	/* if we are connected, we'd better throw out a
	 * deregistration message */
	if (mn.state == MN_CONNECTED) {
		send_registration(REG_DISC);
		DEBUG(DEBUG_MESSAGES, "Deregistration message sent\n");
	}
	if (real_tunnel_up) stop_tunneling();

	/* free session key */
	if (mn.session_key) {
		free(mn.session_key);
		mn.session_key = NULL;
		mn.session_key_len = 0;
	}

	DEBUG(DEBUG_TIMERS, "Clearing TIMER_GEN\n");
	timerclear(&timers[TIMER_GEN]);
        mn.state = MN_DISCONNECTED;
        DEBUG(DEBUG_STATES, "State: Disconnected\n");
	syslog(LOG_INFO, "Disconnected.");
	mn_remove_dynamic_home_addr();
}


/* handle reregistration when the not supported request option can be fixed */
static void reinitiate_state(void)
{
	DEBUG(DEBUG_INFO, "reinitiate_state()\n");

	if (mn.current_adv == NULL) {
		DEBUG(DEBUG_INFO, "reinitiate_state - current_adv == NULL\n");
		mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;
	} else {
		DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: "
		      "%i => %i (reinitiate_state)\n",
		      inet_ntoa(mn.current_adv->addr),
		      mn.current_adv->reg_retry_time,
		      MIN_REGISTRATION_TIME);
		mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
	}
	switch (mn.state) {
	case MN_REQUEST_TUNNEL:
		request_tunnel(STATE_INIT, 0, 0);
		break;
	case MN_CONNECTED:
		reregister();
		break;
	default:
		break;
	}
}


static int check_sol_reply(struct node *node, void *data)
{
	struct agentadv_data *adv;
	struct timeval *tv;
	
	adv = (struct agentadv_data *) node;
	tv = (struct timeval *) data;

	if (timerisset(tv) && cmp_timeval(&adv->last, tv) < 0 &&
	    usec_passed(&adv->last, tv) > MAX_AGENTSOL_REPLY_WAIT) {
		DEBUG(DEBUG_INFO, "  timediff=%u\n",
		      usec_passed(&adv->last, tv));
		DEBUG(DEBUG_INFO, "No reply from MA (addr=%s) to "
		      "agent solicitation within %i usec",
		      inet_ntoa(adv->addr), MAX_AGENTSOL_REPLY_WAIT);
		adv->prio_degrade_percent += SOLICITATION_PRIO_DEGRADE;
		if (adv->prio_degrade_percent > 100)
			adv->prio_degrade_percent = 100;
		DEBUG(DEBUG_INFO, " - prio to %i\n",
		      adv->prio_degrade_percent);
	}

	return TRUE;
}

unsigned int get_average_RTT_to_ha() {
	return get_average_RTT(mn.RTTs);
}
unsigned int get_average_RTT_to_cr(struct ro_entry *cr) {
	return get_average_RTT(cr->RTTs);
}

int set_ha_keepalive_timer_expiration() {
	unsigned int msecs_to_add = get_average_RTT_to_ha();
	DEBUG(DEBUG_TIMERS, "\t - Postpone HA's tunnel keepalive timer expiration by av. RTT: %d msec.\n",
		msecs_to_add);
	return set_timer_expiration(&timers[TIMER_KEEPALIVE], msecs_to_add);
}
int set_ha_reply_wait_timer_expiration() {
	unsigned int msecs_to_add = REPLY_WAIT_TIME;
	DEBUG(DEBUG_TIMERS, "\t - Set HA Reply Wait timer expiration by %d msec.\n", 
		REPLY_WAIT_TIME);
	return set_timer_expiration(&(mn.reg_retry_time), msecs_to_add);
}

int set_cr_keepalive_timer_expiration(struct ro_entry *cr) {
	unsigned int msecs_to_add = get_average_RTT_to_cr(cr);
	DEBUG(DEBUG_TIMERS, "\t - Postpone this CR's tunnel keepalive timer expiration by av. RTT: %d msec.\n",
		msecs_to_add);
	return set_timer_expiration(&(cr->keepalive_timer), msecs_to_add);
}
int set_cr_reply_wait_timer_expiration(struct ro_entry *cr) {
	unsigned int msecs_to_add = REPLY_WAIT_TIME;
	DEBUG(DEBUG_TIMERS, "\t - Set this CR Reply Wait timer expiration by %d msec.\n", 
		REPLY_WAIT_TIME);
	return set_timer_expiration(&(cr->reply_wait_timer), msecs_to_add);
}
static void handle_timeout(void)
{
	struct timeval tv, now, time_waited;
	int i, cr_i, k, new_rtt;
	struct ro_entry *cr;
	
    DEBUG(DEBUG_TIMERS, "[mn.c] handle_timeout - mn.tunnel_up == %d\n",
    	mn.tunnel_up);
	gettimeofday(&tv, NULL);
	gettimeofday(&now, NULL);
	
#ifdef WITH_WIRELESS
	if (config.wlan_ap_poll_interval > -1 &&
	    timerisset(&timers[TIMER_WLAN_AP_POLL]) &&
	    cmp_timeval(&tv, &timers[TIMER_WLAN_AP_POLL]) >= 0) {
		DEBUG(DEBUG_TIMERS, "\tTIMER_WLAN_AP_POLL\n");
		/* poll AP addresses and send agentsol if handoff has
		 * occured (do this for each interface supporting wireless
		 * extensions) */
		monitor_poll_ap_addresses(&mn);
		timers[TIMER_WLAN_AP_POLL].tv_sec = tv.tv_sec;
		timers[TIMER_WLAN_AP_POLL].tv_usec = tv.tv_usec;
		add_usecs(&timers[TIMER_WLAN_AP_POLL],
			  config.wlan_ap_poll_interval);
	}
#endif /* WITH_WIRELESS */
	
	/*------------------------------------------------------------------------------*
	 | Check if a timer is set and the time now >= timer.							|
	 *------------------------------------------------------------------------------*/
	
	/* HA keepalive timer */
	if (timerisset(&timers[TIMER_KEEPALIVE]) &&
	    cmp_timeval(&tv, &timers[TIMER_KEEPALIVE]) >= 0)
	{
		DEBUG(DEBUG_TIMERS, "\tTIMER_KEEPALIVE (HA) expired -> send ping\n");

		/* Send a keepalive to keep the HA tunnel open. */
		send_ping_to_HA();
	}
	/* HA Reply wait timer */
	if (timerisset(&mn.reg_retry_time) &&
	    cmp_timeval(&tv, &mn.reg_retry_time) >= 0)
	{
		DEBUG(DEBUG_TIMERS, "\tReg retry timer (HA) expired -> send new Reg through another eth\n");
		print_trace(mn_traces, "No Reg Reply from HA -> try registering from a different eth\n");
		set_ha_reply_wait_timer_expiration();
		register_through_next_eth();
	}
	/* HA ping failure timers */
	for (i = 0; i < KEEPALIVE_ARRAY_LEN; i++) 
	{
		if (mn.ha_keepalives[i].timer_on &&
			cmp_timeval(&tv, &mn.ha_keepalives[i].failure_timer) >= 0) 
		{
			DEBUG(DEBUG_INFO, "handle_timeout: ping (seq %d) to HA failed. "
				"Ping failure count %d.\n", mn.ha_keepalives[i].seq_number, 
				mn.ping_failure_count + 1);
			
			/* Get the time the ping got waited and init the RTT array with that value */
			/* time_waited = now - sending_time */
			timersub(&now, &mn.ha_keepalives[i].sending_time, &time_waited);
			DEBUG(DEBUG_INFO, "Failed ping was sent %ld.%ld seconds ago\n", 
				time_waited.tv_sec, time_waited.tv_usec);
			new_rtt = time_waited.tv_sec * 1000 + time_waited.tv_usec / 1000;
			insert_RTT(mn.RTTs, new_rtt);
			
			mn.ha_keepalives[i].timer_on = 0;
			mn.ha_keepalives[i].failure_timer.tv_sec = 0;
			mn.ha_keepalives[i].failure_timer.tv_usec = 0;
			mn.ping_failure_count++;
			print_trace(mn_traces, "Lost a ping to HA, fail count %d, av. RTT %d\n", 
				mn.ping_failure_count, get_average_RTT_to_ha());
				
			/* If more than config.max_lost_keepalives failed -> link failure */
			if (mn.ping_failure_count >= config.max_lost_keepalives) 
			{
				print_trace(mn_traces, "Path to HA through %s failed\n",
					mn.device_in_use);
				/* unset keepalive timer */
				timers[TIMER_KEEPALIVE].tv_sec = 0;
				timers[TIMER_KEEPALIVE].tv_usec = 0;
				init_keepalives(mn.ha_keepalives, KEEPALIVE_ARRAY_LEN);
				set_ha_reply_wait_timer_expiration();
				mn.ping_failure_count = 0;
				register_through_next_eth();
			}
		}
	}
	
	/* CRs */
	for (cr_i = 0; cr_i < ro_entry_count; cr_i++) 
	{
		cr = (struct ro_entry *) get_ro_entry_by_index(cr_i);
		
		/* CR keepalive*/
		if (timerisset(&cr->keepalive_timer) &&
		    cmp_timeval(&tv, &cr->keepalive_timer) >= 0)
		{
			DEBUG(DEBUG_TIMERS, "\tCR (%s) KEEPALIVE TIMER expired -> send a keepalive\n", 
				inet_ntoa(cr->cr_hoa));

			/* Add average RTT to the CR keepalive timer. */
			set_cr_keepalive_timer_expiration(cr);
			
			DEBUG(DEBUG_INFO, "handle_timeout: "
				"CR's keepalive timer reset to %li.%06li (in %li sec)\n",
				cr->keepalive_timer.tv_sec, cr->keepalive_timer.tv_usec,
		    	cr->keepalive_timer.tv_sec - time(NULL));
		    			
			/* Send a keepalive to keep the CR tunnel open. */
			send_ping_to_CR(cr);
		}
		
		/* CR binging update (Reregistration) timer */
		if (timerisset(&cr->rereg_timer) &&
		    cmp_timeval(&tv, &cr->rereg_timer) >= 0)
		{
			DEBUG(DEBUG_TIMERS, "\tCR (%s) REREG_TIMER expired -> send a Reg Req\n", 
				inet_ntoa(cr->cr_hoa));
			
			cr->rereg_timer.tv_sec 	= 0;
			cr->rereg_timer.tv_usec = 0;
			
			send_regreq_to_cr(cr, REG_REREG);
		}
		
		/* CR Reply waiting timer */
		if (timerisset(&cr->reply_wait_timer) &&
		    cmp_timeval(&tv, &cr->reply_wait_timer) >= 0)
		{
			print_trace(mn_traces, "No Reg Reply from CR %s for Req of ID %08x %08x "
				"-> try registering from a different eth\n", 
				inet_ntoa(cr->cr_hoa),
				(__u32) ntohl(cr->sent_reg_id[0]), 
				(__u32) ntohl(cr->sent_reg_id[1]));
			DEBUG(DEBUG_TIMERS, "\tCR (%s) REPLY WAITING TIMER expired -> send a new Request\n", 
				inet_ntoa(cr->cr_hoa));

			DEBUG(DEBUG_INFO, "handle_timeout: "
				"CR's keepalive timer reset to %li.%06li (in %li sec)\n",
				cr->reply_wait_timer.tv_sec, cr->reply_wait_timer.tv_usec,
		    	cr->reply_wait_timer.tv_sec - time(NULL));			
			
			cr->ping_failure_count = 0;
			
			if (cr->reg_attempt_count < MAX_REG_ATTEMPTS_TO_CR) {
				cr->reg_attempt_count++;
				register_to_cr_through_next_mr_coa(cr);
			} else {
				print_trace(mn_traces, "Tried %d times to connect CR %s -> remove from RO Cache\n",
					MAX_REG_ATTEMPTS_TO_CR, inet_ntoa(cr->cr_hoa));
				delete_ro_entry(cr);
			}
		}

		/* CR ping failure timers */
		for (k = 0; k < KEEPALIVE_ARRAY_LEN; k++) 
		{	
			if (cr->keepalives[k].timer_on &&
				cmp_timeval(&tv, &cr->keepalives[k].failure_timer) >= 0) 
			{	
				
				/* If this was the initial ping that failed -> try another CR CoA */
				if (cr->rr_state == RR_IN_PROGRESS)
				{
					/* Set the expired timer off */
					cr->keepalives[k].timer_on = 0;
					cr->keepalives[k].failure_timer.tv_sec = 0;
					cr->keepalives[k].failure_timer.tv_usec = 0;
				
					DEBUG(DEBUG_INFO, "Failed to connect to CR CoA %s\n", 
						inet_ntoa(cr->tnl_dst));
					if (cr->tunneling_as_mr) {
						change_cr_tnl_to_next_coa(cr);
						DEBUG(DEBUG_INFO, " - tunneling as MR -> try next CoA: %s\n", 
							inet_ntoa(cr->tnl_dst));
						send_initial_ping_to_CR(cr);
					} else {
						DEBUG(DEBUG_INFO, " - tunneling as CR -> wait for the MR to "
							"init connection\n");
					}
				} else 
				{
					DEBUG(DEBUG_INFO, "handle_timeout: ping (seq %d) to CR (%s) failed. "
							"Ping failure count %d.\n", cr->keepalives[k].seq_number,
							inet_ntoa(cr->cr_hoa), cr->ping_failure_count + 1);
							
					/* Get the time the ping got waited and init the RTT array with that value */
					/* time_waited = now - sending_time */
					timersub(&now, &cr->keepalives[k].sending_time, &time_waited);
					DEBUG(DEBUG_INFO, "Failed ping was sent %ld.%ld seconds ago\n", 
						time_waited.tv_sec, time_waited.tv_usec);
					new_rtt = time_waited.tv_sec * 1000 + time_waited.tv_usec / 1000;
					insert_RTT(cr->RTTs, new_rtt);
				
					/* Set the expired timer off */
					cr->keepalives[k].timer_on = 0;
					cr->keepalives[k].failure_timer.tv_sec = 0;
					cr->keepalives[k].failure_timer.tv_usec = 0;
				
					cr->ping_failure_count++;
					print_trace(mn_traces, "Lost a ping to CR %s, CoA %s, fail count %d, av. RTT %d\n", 
						inet_ntoa(cr->cr_hoa), get_cr_tnl_dst_str(cr), cr->ping_failure_count, 
						get_average_RTT_to_cr(cr));
						
					/* config.max_lost_keepalives pings lost = path failure */
					if (cr->ping_failure_count >= config.max_lost_keepalives) 
					{
						//DEBUG(DEBUG_INFO, "Path to CR %s through %s failed\n",
						//	inet_ntoa(cr->cr_hoa), cr->device_in_use);
						print_trace(mn_traces, "Path to CR %s through %s failed\n",
							inet_ntoa(cr->cr_hoa), cr->device_in_use);
						cr->ping_failure_count = 0;
						/* unset keepalive timer */
						init_keepalives(cr->keepalives, KEEPALIVE_ARRAY_LEN);
						cr->keepalive_timer.tv_sec = 0;
						cr->keepalive_timer.tv_usec = 0;
						handle_cr_path_failure(cr);
					}
				}
			}
		}
	}

	if (timerisset(&timers[TIMER_REREG]) &&
	    cmp_timeval(&tv, &timers[TIMER_REREG]) >= 0)
	{
		DEBUG(DEBUG_TIMERS, "\tTIMER_REREG\n");
		reregister();
	}
	
	if (config.solicitation_interval > -1 &&
	    timerisset(&timers[TIMER_SOLICITATION]) &&
	    cmp_timeval(&tv, &timers[TIMER_SOLICITATION]) >= 0) 
	{
		DEBUG(DEBUG_TIMERS, "\tTIMER_SOLICITATION\n");
		/* check whether some FAs did not reply previous solicitation */
		hashtable_iterator(mn.agentadv, check_sol_reply,
				   &mn.last_scheduled_solicitation);
		/* send agentsol to all interfaces */
		send_solicitations(&mn);
		timers[TIMER_SOLICITATION].tv_sec = tv.tv_sec;
		timers[TIMER_SOLICITATION].tv_usec = tv.tv_usec;
		add_usecs(&timers[TIMER_SOLICITATION],
			  config.solicitation_interval +
			  get_rand32() % MAX_RANDOM_SOLICITATION_DELAY);
		mn.last_scheduled_solicitation = tv;
	}

	if (!timerisset(&timers[TIMER_GEN])) {
		DEBUG(DEBUG_TIMERS, "\tTIMER_GEN is not set\n");
		return;
	}

	if (cmp_timeval(&tv, &timers[TIMER_GEN]) < 0) {
		DEBUG(DEBUG_TIMERS, "\tTIMER_GEN not yet reached - ignoring "
		      "timeout\n");
		return;
	}

	/* disable timer */
	DEBUG(DEBUG_TIMERS, "\tdisabling TIMER_GEN\n");
        timerclear(&timers[TIMER_GEN]);

	switch (mn.state) 
	{
        case MN_DISCONNECTED:
     	   DEBUG(DEBUG_TIMERS, "\tcase: MN_DISCONNECTED\n");
        case MN_AT_HOME:
        	DEBUG(DEBUG_TIMERS, "\tcase: MN_AT_HOME\n");
            break;
        case MN_CLOSE_FOR_HOME:
        	DEBUG(DEBUG_TIMERS, "\tcase: MN_CLOSE_FOR_HOME\n");
            close_for_home(STATE_TIMEOUT);
            break;
        case MN_FIND_AGENT:
        	DEBUG(DEBUG_TIMERS, "\tcase: MN_FIND_AGENT\n");
            find_agent(STATE_TIMEOUT);
			break;
        case MN_PASSIVE_FIND:
        	DEBUG(DEBUG_TIMERS, "\tcase: MN_PASSIVE_FIND\n");
            break;
        case MN_REQUEST_TUNNEL:
        	DEBUG(DEBUG_TIMERS, "\tcase: MN_REQUEST_TUNNEL\n");
        	if (!mn.init_reply_received) bind_ha_tunnel_to_next_iface();
            request_tunnel(STATE_TIMEOUT, 0, 0);
            break;
	    case MN_CONNECTED:
	    	DEBUG(DEBUG_TIMERS, "\tcase: MN_CONNECTED\n");
            break;
		default:
			DEBUG(DEBUG_TIMERS, "\tcase: default\n");
			break;
	}
}


/**
 * from_fd_to_tunnel: Read msg from fd, add MIP TNL header and send the datagram 
 * to dst_ip:dst_port.
 */
int from_fd_to_tunnel(int fd, struct in_addr dst_ip, int dst_port, int socket) {
	char dst_ip_str[INET_ADDRSTRLEN];
	inet_ntop(AF_INET, &dst_ip, dst_ip_str, INET_ADDRSTRLEN);
	int bufsize = 2000;
	char buffer[bufsize];
	int nread;
	
	/*----------------------------------------------------------------------------------*
	 | Leave 4 bytes in the beginning of the buffer and add MIP TNL Msg header there.	|
	 *----------------------------------------------------------------------------------*/
	nread = read(fd, buffer + 4, bufsize - 4);
  	if (nread < 0) {
    	DEBUG(DEBUG_INFO, "[mn.c] from_fd_to_tunnel: Error reading data!\n");
    	return -1;
  	} // else if (nread > ) TODO handle bigger packets?
	
	DEBUG(DEBUG_INFO, " ================>>>> \n");
	DEBUG(DEBUG_INFO, "[mn.c] from_fd_to_tunnel: %d bytes read from tun_fd %d "
		"- add MIP TNL header and send to %s:%d\n", nread, fd, dst_ip_str, ntohs(dst_port));
	
	print_ip_stack(buffer + 4);
	insert_udp_tnl_msg_header(buffer);
	
	return send_udp_to(buffer, nread + 4, dst_ip, dst_port, socket);
}

/* from_buffer_to_tunnel: Add MIP TNL header and send to dst_ip:dst_port.
 */
int from_buffer_to_tunnel(unsigned char *buf, int len, struct in_addr dst_ip, 
		int dst_port, int socket) {
	
	/* Printataan DEBUG-info */
	char dst_ip_str[INET_ADDRSTRLEN];
	char new_buffer[len + 4]; 				// UDP Tunnel Message header + *buf	
	
	inet_ntop(AF_INET, &dst_ip, dst_ip_str, INET_ADDRSTRLEN);
	DEBUG(DEBUG_INFO, " ================>>>> \n");
	DEBUG(DEBUG_INFO, "[mn.c] from_buffer_to_tunnel() - Message from socket %d to %s:%d\n", 
			socket, dst_ip_str, ntohs(dst_port));
	print_ip_stack(buf);
	
	/*--------------------------------------------------------------------------*
	 | Add MIP Tunnel Data Message header into the beginning of the new buffer.	|
	 *--------------------------------------------------------------------------*/
	insert_udp_tnl_msg_header(new_buffer);	
	
	/*------------------------------------------------------------------*
	 | Copy the buffer after the MIP TNL header.						|
	 *------------------------------------------------------------------*/
	memcpy(new_buffer + 4, buf, len);

	return send_udp_to(new_buffer, sizeof(new_buffer), dst_ip, dst_port, socket);
}

/* from_buffer_to_ha_tunnel: Send from buffer to the HA tunnel
 * 
 * @buf: pointer to the buffer containing the datagram
 * @len: length of the datagram in the buffer
 */
int from_buffer_to_ha_tunnel(unsigned char *buf, int len) {
	struct in_addr tnl_dst_ip;						// Tunnel dst address 
	int tnl_dst_port;
	int socket;
	
	tnl_dst_ip.s_addr 	= config.ha_ip_addr.s_addr;	// tunnel to HA
	tnl_dst_port		= htons(config.udp_port);
	socket				= mn.registration_socket;
	
	return from_buffer_to_tunnel(buf, len, tnl_dst_ip, 
		tnl_dst_port, socket);
}

/* from_buffer_to_cr_tunnel: Send from buffer to the CR's tunnel.
 * 
 * @buf: pointer to the buffer containing the datagram
 * @len: length of the datagram in the buffer
 * 
 * If cr->tunneling_as_mr is set
 *  - send from my_reg_port to 434 
 * Else
 *  - send from 434 to distant_reg_port
 * 
 */
int from_buffer_to_cr_tunnel(struct ro_entry *cr, unsigned char *buf, int len) {
	struct in_addr tnl_dst_ip;						// Tunnel dst address 
	int tnl_dst_port;
	int socket;
	
	tnl_dst_ip.s_addr 	= cr->tnl_dst.s_addr;
	
	/* Set src and dst ports corresponding to the tunneling role in this connection */
	if (cr->tunneling_as_mr)
	{
		/* MR to CR: my_reg_port ==> 434 */
		socket				= cr->registration_socket;
		tnl_dst_port		= htons(434);
	} else {	
		/* CR to MR: 434 ==> reg_port */
		socket				= cr->coa_mip_socket;
		tnl_dst_port		= cr->cr_co_port; 	
	}
	return from_buffer_to_tunnel(buf, len, tnl_dst_ip, 
		tnl_dst_port, socket);
}

/* handle_msg_in_ha_tun_fd: Read message from HA's tun_fd, add MIP TNL header
 * and send it to HA's CoA via registration socket.
 */
int handle_msg_in_ha_tun_fd() {
	int fd_to_read_from;
	struct in_addr tnl_dst_ip;
	int tnl_dst_port;
	int out_socket;
	
	fd_to_read_from 	= mn.tun_fd;
	tnl_dst_ip.s_addr	= config.ha_ip_addr.s_addr;
	tnl_dst_port		= htons(config.udp_port);
	out_socket			= mn.registration_socket;
	
	DEBUG(DEBUG_INFO, "Message in HA's tun_fd\n");  
	
	return from_fd_to_tunnel(fd_to_read_from, tnl_dst_ip, tnl_dst_port, out_socket);
}
/* handle_msg_in_cr_tun_fd: Read data from cr->tun_fd and send it 
 * to cr->tnl_dst with ports corresponding to the tunneling roles in
 * this connection:
 * 
 * If cr->tunneling_as_mr
 *   - send from my_reg_port to 434
 * Else 
 *   - send from 434 to distant_reg_port
 * 
 */
int handle_msg_in_cr_tun_fd(struct ro_entry *cr) {
	int fd_to_read_from;
	struct in_addr tnl_dst_ip;
	int tnl_dst_port;
	int socket;

	/*----------------------------------------------------------*
	 | Check if this tunnel is active.							|
	 *----------------------------------------------------------*/
	if (cr->rr_state != RR_ACTIVE) {
		/* Non-active tunnel --> read and discard data */
		int bufsize = 2000;
		char buffer[bufsize];
		int nread = read(cr->tun_fd, buffer, bufsize);
	  	
		DEBUG(DEBUG_INFO, "Message (%d B) in CR's (%s) TUN device not forwarded "
			"because TNL RR state is not active! (state == %d)\n", 
			nread, inet_ntoa(cr->cr_hoa), cr->rr_state);
	
		return -1;
	}
	
	/*----------------------------------------------------------*
	 | Add MIP TNL header and send to CR CoA.					|
	 *----------------------------------------------------------*/
	fd_to_read_from 	= cr->tun_fd;
	tnl_dst_ip.s_addr	= cr->tnl_dst.s_addr;
	
	if (cr->tunneling_as_mr) {
		socket			= cr->registration_socket;
		tnl_dst_port	= htons(434);
	} else {
		socket			= cr->coa_mip_socket;
		tnl_dst_port	= cr->cr_co_port;			/* cr->cr_co_port in nw byte order */			
	}
	DEBUG(DEBUG_INFO, "Message in tun_fd of CR %s\n",
		inet_ntoa(cr->cr_hoa));

	return from_fd_to_tunnel(fd_to_read_from, tnl_dst_ip, tnl_dst_port, socket);
}

int bind_ha_tunnel_to_device(char *dev_name) {
	
	/* Copy the new device name. */
	memcpy(mn.device_in_use, dev_name, IFNAMSIZ);	
	
	/* Bind the registration socket to this device. */
	bind_socket_to_device(mn.registration_socket, mn.device_in_use);
	
	/* Copy eth address into mn.co_addr and mn.local_addr */
	get_iface_address(mn.device_in_use, &mn.local_addr);
	get_iface_address(mn.device_in_use, &mn.co_addr);
	
	/* Initialize RTTs for the new link. */		
	init_RTTs(mn.RTTs, RTT_ARRAY_LEN);

	/* TODO: Listen to the tnl old address for awhile. */
	/* Old tnl listener can not be created to the old device's reg port because the 
	 * mn.registration_socket is originally binded there -> need more sockets
	 * 
	reg_port = get_socket_local_port(mn.registration_socket);
	if (reg_port < 0) {
		DEBUG(DEBUG_INFO, "Could not get HA Registration Socket's local port!\n");
	}
	if (mn.old_tnl_listener > -1) {
		DEBUG(DEBUG_INFO, "Old HA tunnel still up with fd %d -> close\n", mn.old_tnl_listener);
		close(mn.old_tnl_listener);
		mn.old_tnl_listener = -1;
	}
	mn.old_tnl_listener = create_socket(old_iface_addr, reg_port);
	DEBUG(DEBUG_INFO, "Temp old iface listener (fd %d) on %s port %d\n", mn.old_tnl_listener,
		inet_ntoa(old_iface_addr), ntohs(reg_port));
	set_timer_expiration(mn.old_tnl_listener_expiration_time, OLD_HA_TNL_ALIVE_TIME);
	*/
	
	return 0;
}
int bind_ha_tunnel_to_next_iface()
{
	char new_iface[IFNAMSIZ];
	
	DEBUG(DEBUG_INFO, "Update the HAaRO usable nw iface list and send "
		"Registration Request through the next eth in the list.\n");
	get_next_usable_iface(mn.device_in_use, new_iface);
	bind_ha_tunnel_to_device(new_iface);
	DEBUG(DEBUG_INFO, "New device to use in HA tunneling: %s\n", mn.device_in_use);
	return 0;
}
/* register_through_next_eth: Register to HA through the eth device. 
 * 
 * TODO: Listen to the old tnl for awhile 
 */
int register_through_next_eth() {
	bind_ha_tunnel_to_next_iface();
	reregister();
	return 0;
}

int register_to_ha_through_eth(char *iface_name)
{
	if (check_if_iface_usable(iface_name))
	{
		/* Take the new device in use for HA tunneling */
		bind_ha_tunnel_to_device(iface_name);
		DEBUG(DEBUG_INFO, "New device to use in HA tunneling: %s\n", iface_name);	
		reregister();
		return 0;
	} else {
		print_trace(mn_traces, "Could not register to HA through %s because "
			"the iface is invalid for HAaRO!\n", iface_name);
		return -2;
	}
}
/* register_to_cr_through_eth: 
 * 
 * If the RO Entry has a KRm calculated for the new iface (TODO: And the KRm is not expired)   
 *  -> bind the CR connection to the new device and send Reg Request
 *     using this KRm
 * Else
 *  -> bind CoA socket to the new device
 *  -> send HoTI + CoTI
 *  -> calculate KRm when HoT+CoT received
 *  -> bind registration socket etc. to the new device and send Reg Request
 */
int register_to_cr_through_eth(struct ro_entry *cr, char *iface_name)
{
	DEBUG(DEBUG_INFO, "register_to_cr_through_eth: %s\n", iface_name);
	
	/* If the device has KRm (TODO: and the KRm is not expired) 
	 *  -> register through this new device */
	if (start_using_krm_of_iface_if_valid(cr, iface_name))
	{
		print_trace(mn_traces, "KRm already calculated for this iface -> bind and register\n");
		
		/* Take the new device in use */
		bind_cr_connection_to_device(cr, iface_name);
		DEBUG(DEBUG_INFO, "New device to use with CR %s tnl: %s\n", 
			inet_ntoa(cr->cr_hoa), cr->device_in_use);
		return send_regreq_to_cr(cr, REG_CONNECT);
	} else {
		print_trace(mn_traces, "Need a new KRm for the new CoA -> send HoTI + CoTI\n");
		bind_socket_to_device(cr->coti_socket, iface_name);
		memcpy(cr->last_coti_iface, iface_name, IFNAMSIZ);
		go_to_resend_hoti_coti_through_new_iface_state(cr);
		return 0;
	}
}
/* check_and_bind_cr_tunnel_to_eth: Check if the iface_name is a valid
 * iface and bind the tunnel to the CR to that iface.  
 */
int check_and_bind_cr_tunnel_to_eth(struct ro_entry *cr, char *iface_name)
{
	if (check_if_iface_usable(iface_name))
	{
		register_to_cr_through_eth(cr, iface_name);
		return 0;	
	} else {
		print_trace(mn_traces, "Could not register to CR %s through %s because "
			"the iface is invalid for HAaRO!\n", inet_ntoa(cr->cr_hoa), iface_name);
		return -2;
	}
}

/* register_cr_through_next_eth: Find out the next eth in the 
 * usable_nw_ifaces list and register (or send HoTI + CoTI)
 * from it.
 */
int register_to_cr_through_next_mr_coa(struct ro_entry *cr) {
	int i;
	
	DEBUG(DEBUG_INFO, "Update the HAaRO usable nw iface list and send "
		"Registration Request through the next eth in the list.\n");
	//update_usable_nw_ifaces();
	
	/* Browse all nw device usable in HAaRO */
	for (i = 0; i < usable_nw_iface_count; i++) 
	{
		/* Find out which device in the array is currently in use and register through the next one */
		if (strncmp(usable_nw_ifaces[i].dev_name, cr->device_in_use, IFNAMSIZ) == 0) 
		{
			DEBUG(DEBUG_INFO, "Device currently in use: [%d] %s\n", i, cr->device_in_use);
			
			/* Increase i, except if it's in the last index */
			if (i != (usable_nw_iface_count - 1)) 
				i++;
			else 
				i = 0;
			
			return register_to_cr_through_eth(cr, usable_nw_ifaces[i].dev_name);
		}
	}
	
	/* Device currently in use not found in the list -> take the first in the list */
	DEBUG(DEBUG_INFO, "Device currently in use (%s) no more usable! -> "
		"Start using the first in the list.\n", cr->device_in_use);
	
	return register_to_cr_through_eth(cr, usable_nw_ifaces[0].dev_name);
}

int ping_next_cr_coa(struct ro_entry *cr) {
	DEBUG(DEBUG_INFO, "Try pinging the next CoA of CR %s\n", inet_ntoa(cr->cr_hoa));
	DEBUG(DEBUG_INFO, " - current CoA %s\n", inet_ntoa(cr->tnl_dst));
	
	change_cr_tnl_to_next_coa(cr);
	DEBUG(DEBUG_INFO, " - next CoA %s\n", inet_ntoa(cr->tnl_dst));
	
	return send_ping_to_CR(cr);
}
/* handle_cr_path_failure: Tunnel to this CR failed. Send a Registration 
 * Request from a different MR CoA to the CR HoA. 
 */
int handle_cr_path_failure(struct ro_entry *cr) {
	if (cr->tun_fd > 0) 
	{
		close(cr->tun_fd);
		cr->tun_fd = -1;
		cr->tnl_count--;
	}
	cr->rr_state = RR_INACTIVE;
	print_trace(mn_traces, "tnl to CR %s inactive\n", inet_ntoa(cr->cr_hoa));	
	return register_to_cr_through_next_mr_coa(cr);
	
}

/* set_ping_timer: Fill the first free element in the ping info array 
 * with sequence number, sending time and failure timer that expires in 
 * config.keepalive_wait_percentage * average_RTT.
 */
int set_ping_timer(struct requested_ping *ping_array, int seq_number) {
	int i;
	
	/* Look for a free element in the array */
	for (i = 0; i < KEEPALIVE_ARRAY_LEN; i++) {
		if (!ping_array[i].timer_on) {
			
			ping_array[i].seq_number = seq_number;
			ping_array[i].timer_on = 1;
			
			/* failure_timer = now + p * average_RTT */
			gettimeofday(&ping_array[i].failure_timer, NULL);
			add_usecs(&ping_array[i].failure_timer, 
				config.keepalive_wait_percentage * get_average_RTT_to_ha() * 1000);
			
			gettimeofday(&ping_array[i].sending_time, NULL);
			
			return 0;
		}
	}
	DEBUG(DEBUG_INFO, "set_ping_timer: No free elements in ping_array!\n");
	return -1;
}

/* send_ping_to_HA: Send ping through the tunnel from MR CoA to HA and add 
 * KEEPALIVE_INTERVAL to the keepalive_timer.
 * 
 * Protocol stack
 *  ICMP: type ICMP Echo Request
 *  IP	: src MR HoA, dst HA
 *  MIP	: type TNL
 *  UDP	: src rand(), dst 434
 *  IP	: src MR CoA, dst HA
 * 
 */
int send_ping_to_HA() {	
	int i, ip_hdr_size, total_packet_size;
	unsigned char buffer[84];
	struct iphdr *iph;
	struct icmphdr *icmph;
	unsigned char *icmp_data_ptr;
	unsigned short icmp_sequence;
	unsigned char icmp_type = 8;	// Echo Request
	struct timeval now;
	char icmp_data[56];
	int icmp_data_len = 56;
	struct in_addr tunneled_ip_src; 	// Addresses of the IP inside the tunnel
	struct in_addr tunneled_ip_dst;
	struct in_addr tnl_dst_ip;			// Tunnel dst address 
	int tnl_dst_port;
	
	tunneled_ip_src.s_addr = config.mn_home_ip_addr.s_addr;	// ICMP from MR HoA 
	tunneled_ip_dst.s_addr = config.ha_ip_addr.s_addr;		// ICMP to HA 

	tnl_dst_ip.s_addr 	= config.ha_ip_addr.s_addr;			// tunnel to HA
	tnl_dst_port		= htons(config.udp_port);
	
	/* Get current time and place it in the ICMP data field */
	gettimeofday(&now, NULL);
	
	/*------------------------------------------------------------------*
	 | Find out the ICMP sequence number.								|
	 *------------------------------------------------------------------*/	
	icmp_sequence = get_new_icmp_seq();
	
	/*--------------------------------------------------------------*
	 | Generate ICMP data											|
	 *--------------------------------------------------------------*/
	memcpy(icmp_data, &now.tv_sec, sizeof(long int));
	memcpy(icmp_data + sizeof(long int), &now.tv_usec, sizeof(long int));
	
	/*--------------------------------------------------------------*
	 | Add IP + ICMP headers in the buffer							|
	 *--------------------------------------------------------------*/
	total_packet_size = fill_icmp_req_in_ip(buffer, tunneled_ip_src, tunneled_ip_dst, 
				icmp_sequence, icmp_type, icmp_data, icmp_data_len);

	/*--------------------------------------------------------------------------*
	 | Set the ping timer to expire if no reply for this ping gets received.	|
	 *--------------------------------------------------------------------------*/
	set_ping_timer(mn.ha_keepalives, icmp_sequence);
	set_ha_keepalive_timer_expiration();
	
	DEBUG(DEBUG_INFO, "[mn.c] send_ping_to_HA: %s, seq %d, sending time %li.%06li\n",
		inet_ntoa(tnl_dst_ip), icmp_sequence, now.tv_sec, now.tv_usec);

	/*------------------------------------------------------------------*
	 | Send ICMP in IP packet to the HA via tunnel.						|
	 *------------------------------------------------------------------*/
	/* Could also be written in tun_fd so that it would be noticed in select but it's faster this way */
	return from_buffer_to_tunnel(buffer, sizeof(buffer), tnl_dst_ip, tnl_dst_port, mn.registration_socket);
}

/* send_ping_to_CR: Send ping through the tunnel from MR CoA to CR CoA.
 * 
 * Protocol stack
 *  ICMP
 *  IP	: src MR HoA, dst CR HoA
 *  MIP	: type TNL
 *  UDP	: src rand(), dst 434
 *  IP	: src MR CoA, dst CR CoA
 * 
 */
int send_ping_to_CR(struct ro_entry *cr) {
	
	unsigned char buffer[84];
	int icmp_data_len = 56;
	unsigned char icmp_data[icmp_data_len];
	unsigned char icmp_type = 8;	// Echo Request
	unsigned short icmp_seq;
	struct in_addr tunneled_ip_src; // Addresses of the IP inside the tunnel
	struct in_addr tunneled_ip_dst;
	int icmp_packet_size;
	char cr_hoa_str[INET6_ADDRSTRLEN];
	char cr_coa_str[INET6_ADDRSTRLEN];
	
	assert(cr != NULL);
	
	icmp_seq 					= get_new_icmp_seq();
	
	/* IP addresses of the ICMPinIP packet inside the tunnel */
	tunneled_ip_src.s_addr = config.mn_home_ip_addr.s_addr;
	tunneled_ip_dst.s_addr = cr->cr_hoa.s_addr;

	inet_ntop(AF_INET, &cr->cr_hoa, cr_hoa_str, INET6_ADDRSTRLEN);
	inet_ntop(AF_INET, &cr->tnl_dst, cr_coa_str, INET6_ADDRSTRLEN);
	
	DEBUG(DEBUG_INFO, "send_ping_to_CR: CR %s, CoA %s, seq %d\n",
		cr_hoa_str, cr_coa_str, icmp_seq); 

	strncpy(icmp_data, "This is a ping to CR", icmp_data_len);
	
	/*--------------------------------------------------------------*
	 | Add IP + ICMP headers in the buffer							|
	 *--------------------------------------------------------------*/
	icmp_packet_size = fill_icmp_req_in_ip(buffer, tunneled_ip_src, tunneled_ip_dst, 
				icmp_seq, icmp_type, icmp_data, icmp_data_len);
	
	//gettimeofday(&cr->ping_sending_time, NULL); No more in use
		
	/*--------------------------------------------------------------------------*
	 | Set the ping timer to expire if no reply for this ping gets received.	|
	 *--------------------------------------------------------------------------*/
	set_ping_timer(cr->keepalives, icmp_seq);
	
	/*--------------------------------------------------------------*
	 | Write the ICMP request into this CR CoA's tun fd				|
	 *--------------------------------------------------------------*/
	return from_buffer_to_cr_tunnel(cr, buffer, sizeof(buffer));
}

/* send_initial_ping_to_CR: Put the RR State to IN_PROGRESS and send a ping
 * to the CR. */
int send_initial_ping_to_CR(struct ro_entry *cr) {
	cr->rr_state = RR_IN_PROGRESS;
	print_trace(mn_traces, "tnl to CR %s in progress\n", inet_ntoa(cr->cr_hoa));
	return send_ping_to_CR(cr);
}

int main(int argc, char *argv[])
{
	int n, i, j, usec, hoa_dev_tun_fd;
	fd_set set;
	struct timeval tv;
	struct timeval now;
	struct in_addr dummy_addr;
	
	dummy_addr.s_addr = 0;
	srand(time(NULL));
	
	DEBUG(DEBUG_INFO, "Note that the system will not work if there is a default route with metric 0.\n");

	memset(&mn, 0, sizeof(mn));
	mn.info_str = "initialization in progress";

	program_name = parse_long_options(argc, argv, "mobile node",
					     PACKAGE, VERSION, dynamics_usage);

	mn_parse_command_line(argc, argv);

	if (!mn_init()) {
		fprintf(stderr, "Initialization failed.\n");
		exit(0);
	}
	
	/*--------------------------------------------------------------*
	 | Some extra initializations to support HAaRO.					|
	 *--------------------------------------------------------------*/
	for (i = 0; i < NW_IFACE_ARRAY_LEN; i++) usable_nw_ifaces[i].listener_socket = -1;
	dynamics_strlcpy(mn.hoa_dev, "HoA_DEVICE", sizeof(mn.hoa_dev));
	dyn_ip_tunnel_add_udpip(mn.hoa_dev, dummy_addr, dummy_addr, &hoa_dev_tun_fd);
	timers[TIMER_KEEPALIVE].tv_sec = 0;
	timers[TIMER_KEEPALIVE].tv_usec = 0;
	
	/* Find out the device in use and it's index in mn.iface array */
	dyn_ip_route_get(config.ha_ip_addr, mn.device_in_use, IFNAMSIZ);
	for (j = 0; j < mn.device_count; j++) {
		if (strncmp(mn.iface[j].device, mn.device_in_use, IFNAMSIZ) == 0) {
			mn.dev_index_in_iface_array = j;
			DEBUG(DEBUG_INFO, "mn.dev_index_in_iface_array = %d\n", j);
			break;	
		}
		if (j == mn.device_count - 1) 
			DEBUG(DEBUG_INFO, "mn.dev_index_in_iface_array not found!!!\n");
	}
	/* Copy eth address into mn.co_addr and mn.local_addr */
	get_iface_address(mn.device_in_use, &(mn.local_addr));
	get_iface_address(mn.device_in_use, &(mn.co_addr));
	
	init_RTTs(mn.RTTs, RTT_ARRAY_LEN); 		/* Set the 5 latest RTTs to be 2000 ms each. */
	
	init_keepalives(mn.ha_keepalives, KEEPALIVE_ARRAY_LEN);
	
	print_trace(mn_traces, "MR initialized - using iface %s, CoA %s\n", mn.device_in_use,
		inet_ntoa(mn.co_addr));
			
	DEBUG(DEBUG_INFO, "--------------------------------------------\n");
	DEBUG(DEBUG_INFO, "MR local addr:    \t\t%s\n", inet_ntoa(mn.local_addr));
	DEBUG(DEBUG_INFO, "MR CoA:           \t\t%s\n",	inet_ntoa(mn.co_addr));
	DEBUG(DEBUG_INFO, "Nw device in use: \t\t%s\n",	mn.device_in_use);
	DEBUG(DEBUG_INFO, "NEMO list: ");
	struct nemo_nw_entry *nemo_nw = (struct nemo_nw_entry *) list_get_first(&config.nemo_nw_list);
	if (nemo_nw == NULL) {
		DEBUG(DEBUG_INFO, "empty.\n");
	} else {
		while (1) { /* Loop until an empty node */
			DEBUG(DEBUG_INFO, "\t\t\t%s/%d", inet_ntoa(nemo_nw->prefix), nemo_nw->prefix_length);
			nemo_nw = (struct nemo_nw_entry *) list_get_next((struct node *) nemo_nw);
			if (nemo_nw == NULL) break;
			DEBUG(DEBUG_INFO, "\n\t");
		}
		DEBUG(DEBUG_INFO, "\n");
	}
	
	DEBUG(DEBUG_INFO, "Initial RTT:               \t%d msec\n", config.init_rtt);
	DEBUG(DEBUG_INFO, "Max lost keepalives:       \t%d\n", 		config.max_lost_keepalives);
	DEBUG(DEBUG_INFO, "Keepalive wait percentage: \t%f\n", 		config.keepalive_wait_percentage * 100);
	DEBUG(DEBUG_INFO, "Min keepalive interval:    \t%d msec\n", config.min_keepalive_interval);
	update_usable_nw_ifaces();
	DEBUG(DEBUG_INFO, "--------------------------------------------\n");
	
	print_trace(mn_traces, "System's networking devices to be used in tunneling:\n");
	for (i = 0; i < usable_nw_iface_count; i++)
	{
		print_trace(mn_traces, " - %s %s\n", usable_nw_ifaces[i].dev_name, 
			inet_ntoa(usable_nw_ifaces[i].ip_addr));
	}	
	
	mn.node_key[0] = (__u64) rand() << 32 | (__u64) rand();
	mn.node_key[1] = (__u64) rand() << 32 | (__u64) rand();

	mn.current_nonce_index = -1;
	create_new_nonce();
	
	DEBUG(DEBUG_INFO, "Created the first nonce: %d. %llu\n", 
		mn.current_nonce_index, mn.nonces[0]);
		
	mn.tun_fd 				= 0;
	mn.icmp_seq_num			= 1;
	mn.use_auth_alg			= 4;					/* jpaaso - Hard coded because caused errors */
	mn.tunnel_mode 			= API_TUNNEL_FULL_HA;	/* jpaaso - Use no FAs */
	bind_socket_to_device(mn.registration_socket, mn.device_in_use);
	mn.old_tnl_listener 	= -1;
	mn.init_reply_received 	= 0;


	if (mn.opt_connect) {
		/* try to connect to a FA if MN is using FA decapsulation (i.e.
		 * its own home IP addr also in the foreign network) or if the
		 * home address is not assigned to a local interface with MN
		 * decapsulation */
		mn.info_str = "trying to connect";

		/* jpaaso - is_coloc_addr_foreign() returns incorrectly 0 so the check is disabled
		if (config.enable_fa_decapsulation || is_coloc_addr_foreign())
		{*/
		
		/* jpaaso - Always start by finding the HA */
		find_agent(STATE_INIT);
		
		//} else at_home();
	}
	
	/*--------------------------------------------------------------*
	 | MN main loop.												|
	 *--------------------------------------------------------------*/
	while (mn.state != MN_STOP) 
	{
		check_old_tunnel_expiration();	
		FD_ZERO(&set);
		
		gettimeofday(&now, NULL);
				
		if (mn.old_tnl_listener > -1)
		{
			/* If now is later than the old HA tnl expiration time */
			if (cmp_timeval(&now, &mn.old_tnl_listener_expiration_time) >= 0)
			{
				DEBUG(DEBUG_INFO, "Old HA tunnel listener (fd %d) expired -> close\n", mn.old_tnl_listener);
				close(mn.old_tnl_listener);
				mn.old_tnl_listener = -1;
			} else {
				DEBUG(DEBUG_INFO, "HA tunnel (fd %d) into fd_set\n", mn.old_tnl_listener);
				FD_SET(mn.old_tnl_listener, &set);
			}
		}
		
		/*------------------------------------------------------------------------------*
		 | Check the RO Cache's triggers and put the Cache's sockets into the fd_set	|
		 *------------------------------------------------------------------------------*/
		handle_ro_cache_triggers_and_sockets(&set);
		
	  	/* Print HA's RTT array */
	  	DEBUG(DEBUG_INFO, "RTTs to HA: [ ");
	  	for (i = 0; i < RTT_ARRAY_LEN; i++) {
			DEBUG(DEBUG_INFO, "%d ", mn.RTTs[i]);
	  	}
		DEBUG(DEBUG_INFO, "]\n");		
		/*------------------------------------------------------------------*
		 | Check if a new nonce needs to be created							|
		 *------------------------------------------------------------------*/
		if (time(NULL) - mn.current_nonce_creation_time.tv_sec >= NONCE_CREATE_INTERVAL)
		{
			create_new_nonce();
		}

		FD_SET(mn.registration_socket, &set);
		FD_SET(mn.hoa_listener_socket, &set);			/* HoA:434 */
		FD_SET(mn.hoa_socket, &set);					/* HoA:rand (for receiving HoT) */
		
		/* Port 434 listeners for all ifaces into the fd_set */
		for (i = 0; i < NW_IFACE_ARRAY_LEN; i++) {
			if (usable_nw_ifaces[i].listener_socket > -1)
				FD_SET(usable_nw_ifaces[i].listener_socket, &set);
		}
		for (i = 0; i < MAX_INTERFACES; i++) {
			if (mn.iface[i].s_adv > -1)
				FD_SET(mn.iface[i].s_adv, &set);
		}
		if (mn.api_rw_socket > -1)
			FD_SET(mn.api_rw_socket, &set);
		if (mn.api_ro_socket > -1)
			FD_SET(mn.api_ro_socket, &set);
#ifdef INCLUDE_IPAY
		FD_SET(mn.ipay_sock, &set);
		FD_SET(mn.ipay_sock_fa, &set);
#endif
		if (mn.rtnetlink_socket > -1)
			FD_SET(mn.rtnetlink_socket, &set);
		
		/*------------------------------------------------------------------*
	 	 | Add the tun_fds of HA and CRs in the fd_set						|
	  	 *------------------------------------------------------------------*/
	  	/* Catches messages from the kernel to the tunnel etc. */
	  	add_tun_fds_into_fd_set(&set);
	  	
		/*--------------------------------------------------------------*
	 	 | select														|
	  	 *--------------------------------------------------------------*/
	  	/* Get the next timeval select needs to return at. */  	
		if (get_next_timeout(&tv)) 
		{	
			DEBUG(DEBUG_TIMERS, "Current time:             \t%ld.%ld\n", 
				now.tv_sec, now.tv_usec);
			DEBUG(DEBUG_TIMERS, "HA Keepalive timer set to \t%ld.%ld\n", 
				timers[TIMER_KEEPALIVE].tv_sec, timers[TIMER_KEEPALIVE].tv_usec);
			DEBUG(DEBUG_TIMERS, "HA Rereg timer set to     \t%ld.%ld\n",
				timers[TIMER_REREG].tv_sec, timers[TIMER_REREG].tv_usec);
			DEBUG(DEBUG_INFO, "Next timeout                \t%ld.%ld\n", 
				tv.tv_sec, tv.tv_usec);
			
			n = select(FD_SETSIZE, &set, NULL, NULL, &tv);
		} else {
			DEBUG(DEBUG_INFO, "No timeouts to select!\n");
			n = select(FD_SETSIZE, &set, NULL, NULL, NULL);
		}
		DEBUG(DEBUG_INFO, "select returned %d\n", n);
		
		if (n == 0) {
			/* select returned because the timer tv expired */
			handle_timeout();
			continue;
		} else if (n < 0) {
		        switch (errno) {
				case EINTR:
			        break;
				default:
			        LOG2(LOG_ERR, "main - select: %s\n",
				     strerror(errno));
				/* this used to exit dynmnd with
				 * mn.state = MN_STOP;
				 * but this might be a temporary problem, so
				 * try agains; use a small delay to prevent
				 * hogging all the CPU in case of select()
				 * returning error all the time */
				sleep(1);
			}
			continue;
		}		
		/* select returned because file descriptor(s) got set */
		
		/*------------------------------------------------------------------*
	 	 | Check if one of the tun_fds is set								|
	  	 *------------------------------------------------------------------*/
	  	check_if_ro_fds_set(&set);
	  	
	  	/*------------------------------------------------------------------*
	 	 | Listen to Reg and TNL messages									|
	  	 *------------------------------------------------------------------*/
		if (FD_ISSET(mn.registration_socket, &set)) {
		    i = handle_registration(mn.registration_socket);
		    if (i > 0) {
				/* try to reregister */
				reinitiate_state();
		    }
		}
		if (mn.old_tnl_listener >= 0)
		{
			if (FD_ISSET(mn.old_tnl_listener, &set))
				DEBUG(DEBUG_INFO, "Old HA registration socket %d is set\n", mn.old_tnl_listener);
				handle_registration(mn.old_tnl_listener);
		}	
		
		for (i = 0; i < MAX_INTERFACES; i++) {
			if (mn.iface[i].s_adv > -1 &&
			    FD_ISSET(mn.iface[i].s_adv, &set))
			{
				handle_icmp(&mn.iface[i]);
			}
		}

		if (mn.rtnetlink_socket > -1 &&
		    FD_ISSET(mn.rtnetlink_socket, &set)) {
			int res = dyn_ip_monitor_get(
				mn.rtnetlink_socket,
				(config.enforce_routes &&
				 mn.state != MN_DISCONNECTED) ?
				&mn.cur_route_info : NULL);
			if (res == 1 || res == 2)
				check_interfaces(mn.iface, MAX_INTERFACES);
		}
		
		/*------------------------------------------------------------------*
	 	 | Check if messages from the API.									|
	  	 *------------------------------------------------------------------*/
		if (mn.api_rw_socket > -1 && FD_ISSET(mn.api_rw_socket, &set))
			handle_api(mn.api_rw_socket, 1);

		if (mn.api_ro_socket > -1 && FD_ISSET(mn.api_ro_socket, &set))
			handle_api(mn.api_ro_socket, 0);

#ifdef INCLUDE_IPAY
		if (FD_ISSET(mn.ipay_sock, &set))
			handle_ipay(mn.ipay_sock);
		if (FD_ISSET(mn.ipay_sock_fa, &set))
			handle_ipay(mn.ipay_sock_fa);
#endif

		/* check expired agent advs. If current FA's agent
                   adv. has expired move to FIND_AGENT state */
		if (check_expired_agent_advs())
			find_agent(STATE_INIT);
		
		for (i = 0; i < FD_SETSIZE; i++)
		{
			if ( FD_ISSET( i, &set ) )
				DEBUG(DEBUG_INFO, "FD %d is set\n", i );
		}
		//usleep(1000000);
	}
		
	clean_up(0);
	return 0;
}
