/* $Id: message.h,v 1.59 2001/08/23 17:00:47 jm Exp $
 * Message structures and code defines for agent advertisment and registration
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifndef MESSAGE_H
#define MESSAGE_H

#include <netinet/in.h>

#include "owntypes.h"

#ifdef OS_AMIGAOS

/* AMIGAOS */

#define ATTRIBUTE_PACKED

#else

/* default - Linux */

#define ATTRIBUTE_PACKED __attribute__ ((packed))

#endif


/* Vendor extension defines (RFC 3115) */

/* Critical Vendor/Organization Specific Extension */
#define VENDOR_EXT_TYPE1 38
#define VENDOR_EXT_HEADER1 __u8 type; __u8 reserved; __u16 length; \
			   __u32 vendor_id; __u16 sub_type;

/* Normal Vendor/Organization Specific Extension (NVSE) */
#define VENDOR_EXT_TYPE2 134
#define VENDOR_EXT_HEADER __u8 type; __u8 length; __u16 reserved; \
			  __u32 vendor_id; __u16 sub_type;
#define NVSE_HEADER_LEN 8

/* SMI Network Management Private Enterprise Codes:
 * 5202  Helsinki University of Technology/CS-laboratory */
#define VENDOR_ID_DYNAMICS 5202

#define VENDOR_EXT_VERSION 1

/* sub types for Dynamics vendor extensions */
#define VENDOR_EXT_DYNAMICS_AGENTADV 1
#define VENDOR_EXT_DYNAMICS_OPTIONS 2
#define VENDOR_EXT_DYNAMICS_FA_KEYREQ 3
#define VENDOR_EXT_DYNAMICS_FA_PUBKEY 4
#define VENDOR_EXT_DYNAMICS_MN_KEYREQ 5
#define VENDOR_EXT_DYNAMICS_MN_KEYREP 6
#define VENDOR_EXT_DYNAMICS_FA_KEYREP 7
#define VENDOR_EXT_DYNAMICS_FA_PUBKEYREP 8
#define VENDOR_EXT_DYNAMICS_FF_AUTH 9
#define VENDOR_EXT_DYNAMICS_SK_AUTH 10
#define VENDOR_EXT_DYNAMICS_PUBKEY_HASH 11
#define VENDOR_EXT_DYNAMICS_FA_NAI 12
#define VENDOR_EXT_DYNAMICS_PREVIOUS_FA_NAI 13
#define VENDOR_EXT_DYNAMICS_GRE_KEY 14
#define VENDOR_EXT_DYNAMICS_SFA_DEBUG 15
#define VENDOR_EXT_DYNAMICS_PRIV_HA 16
#define VENDOR_EXT_DYNAMICS_SHA_HA_AUTH 17
#define VENDOR_EXT_DYNAMICS_NONCE 18

struct vendor_ext_header {
	VENDOR_EXT_HEADER
} ATTRIBUTE_PACKED;


/* ICMP messages */

struct router_adv {
	__u8 type;       /* always 9 */
	__u8 code;
	__u16 checksum;
	__u8 num_addr;
	__u8 entry_size; /* always 2 */
	__u16 lifetime;
} ATTRIBUTE_PACKED;

/* struct router_adv is followed by num_addr * struct router_adv_router */
struct router_adv_router {
	__u32 router_addr;
	__u32 pref_level;
} ATTRIBUTE_PACKED;

/* RFC 2002, sec. 2.1.1 */
#define AGENT_ADVERTISEMENT_EXTENSION_TYPE 16

struct agent_adv_ext {
	__u8 type;    /* always 16 */
	__u8 length;
	__u16 seq;
	__u16 reg_lifetime;
	__u16 opts;
} ATTRIBUTE_PACKED;

/* struct agent_adv_ext is followed by
 * N '__u32 co_addr's (length = 6 + 4*N)
 */

/* advertisement extension options
 * RFC 2002, sec. 2.1.1 (flags R .. V)
 */
#define AGENT_ADV_REGISTRATION_REQUIRED 0x8000
#define AGENT_ADV_BUSY 0x4000
#define AGENT_ADV_HOME_AGENT 0x2000
#define AGENT_ADV_FOREIGN_AGENT 0x1000
#define AGENT_ADV_MINIMAL_ENCAPSULATION 0x0800
#define AGENT_ADV_GRE_ENCAPSULATION 0x0400
#define AGENT_ADV_VJ_HDR_COMPRESSION 0x0200
/* RFC 3024, sec. 3.1 (flag T) */
#define AGENT_ADV_BIDIR_TUNNELING 0x0100

/* RFC 2002, sec. 2.1.2 */
#define PREFIX_LENGTHS_EXTENSION_TYPE 19

struct prefix_len_ext {
	__u8 type;    /* always 19 */
	__u8 length;
	/* followed by length 'Prefix Length' values (each one byte) in the
	 * order that the router adddresses are listed in router adv. */
} ATTRIBUTE_PACKED;


#define MIN_AGENT_ADV_DYNAMICS_LEN (NVSE_HEADER_LEN + 3)
/* Own vendor extension for agent advertisements */
struct agent_adv_dynamics {
	VENDOR_EXT_HEADER
	__u8 version; /* vendor extension version */
	__u8 opts;
	__u8 reserved2;
#ifdef INCLUDE_IPAY
	__u32 timePrice;
	__u32 bytePrice;
#endif
} ATTRIBUTE_PACKED;

/* agent_adv_dynamics - opts: */
#define AGENT_ADV_OWN_TRIANGLE_TUNNELING 0x80


struct agent_sol {
	__u8 type;  /* always 10 */
	__u8 code;  /* always 0 */
	__u16 checksum;
	__u32 reserved; /* always 0 */
} ATTRIBUTE_PACKED;

/* the timestamp in the id fields uses NTP time format and is in seconds
 * since 00:00 (midnight) 1 January 1900; i.e. we must add 70 years
 * (including the leap days) to the UNIX time stamp (seconds since beginning
 * of 1970) */
#define UNIX_NTP_DIFF 2208988800U

/*
 * UDP messages
 */
#define REG_REQ 1
#define REG_REQ_ID_LEN 8

/* RFC 2002, sec. 3.3 */
#define REGREQ_SIMULTANEOUS_BINDINGS 0x80
#define REGREQ_BROADCAST_DATAGRAMS   0x40
#define REGREQ_MN_DECAPS             0x20
#define REGREQ_MINIMAL_ENCAPS        0x10
#define REGREQ_GRE_ENCAPS            0x08
#define REGREQ_VJ_HC                 0x04
/* RFC 3024, sec. 3.2 */
#define REGREQ_REVERSE_TUNNEL        0x02
#define REGREQ_RESERVED              0x01

/* Registration Request */
struct reg_req {
	__u8 type;
	__u8 opts;
	__u16 lifetime;
	struct in_addr home_addr;
	struct in_addr ha_addr;
	struct in_addr co_addr;
	__u32 id[2];  /* timestamp */
} ATTRIBUTE_PACKED;

#define REG_REP 3

/* Registration reply codes - accepted */
/* RFC 2002 */
#define REGREP_ACCEPTED              0
#define REGREP_ACCEPTED_NO_SB        1
#define REGREP_PREACCEPTED           200	/* jpaaso - for HAaRO */


/* Registration reply codes - denied by FA */
/* RFC 2002 */
#define REGREP_REASON_UNSPEC_FA     64
#define REGREP_ADMIN_PROHIBITED_FA  65
#define REGREP_NO_RESOURCES_FA      66
#define REGREP_MN_FAILED_AUTH_FA    67
#define REGREP_HA_FAILED_AUTH_FA    68
#define REGREP_LONG_LIFETIME_FA     69
#define REGREP_BAD_REQUEST_FA       70
#define REGREP_BAD_REPLY_FA         71
#define REGREP_ENCAP_UNAVAIL_FA     72
#define REGREP_VJ_UNAVAIL_FA        73
/* RFC 3024, sec. 3.4 */
#define REGREP_REVERSE_TUNNEL_UNAVAIL_FA 74
#define REGREP_REVERSE_TUNNEL_MANDATORY_FA 75
#define REGREP_REVERSE_TUNNEL_MN_TOO_DISTANT_FA 76
/* rfc2002-bis-02 */
#define REGREP_INVALID_CAREOF_FA    77
#define REGREP_REGISTRATION_TIMEOUT_FA 78
/* RFC 3024, sec. 3.4 */
#define REGREP_DELIVERY_STYLE_NOT_SUPPORTED_FA 79
/* RFC 2002 */
#define REGREP_HN_UNCREACHABLE_FA   80
#define REGREP_HA_HOST_UNCREACHABLE_FA 81
#define REGREP_HA_PORT_UNCREACHABLE_FA 82
#define REGREP_HA_UNREACHABLE_FA    88
/* RFC 2794 */
#define REGREP_NONZERO_HOMEADDR_REQD_FA 96
#define REGREP_MISSING_NAI_FA 97
#define REGREP_MISSING_HOME_AGENT_FA 98
#define REGREP_MISSING_HOMEADDR_FA 99
/* RFC 3115 */
#define REGREP_UNSUPP_VENDOR_ID_MN_FA 100
#define REGREP_UNSUPP_VENDOR_ID_HA_FA 101
/* RFC 3012 */
#define REGREP_UNKNOWN_CHALLENGE_FA 104
#define REGREP_MISSING_CHALLENGE_FA 105
#define REGREP_STALE_CHALLENGE_FA 106
/* draft-ietf-mobileip-aaa-key-07.txt */
#define REGREP_MISSING_MN_FA_FA 107

/* Registration reply codes - denied by HA */
/* RFC 2002 */
#define REGREP_REASON_UNSPEC_HA    128
#define REGREP_ADMIN_PROHIBITED_HA 129
#define REGREP_NO_RESOURCES_HA     130
#define REGREP_MN_FAILED_AUTH_HA   131
#define REGREP_FA_FAILED_AUTH_HA   132
#define REGREP_ID_MISMATCH_HA      133
#define REGREP_BAD_REQUEST_HA      134
#define REGREP_TOO_MANY_SIMULTANEOUS 135
#define REGREP_UNKNOWN_HA_HA       136
/* RFC 3024, sec. 3.4 */
#define REGREP_REVERSE_TUNNEL_UNAVAIL_HA 137
#define REGREP_REVERSE_TUNNEL_MANDATORY_HA 138
#define REGREP_ENCAP_UNAVAIL_HA 139
/* RFC 3115 */
#define REGREP_UNSUPP_VENDOR_ID_MN_HA 140
#define REGREP_UNSUPP_VENDOR_ID_FA_HA 141

/* jpaaso - added REGREP_PREACCEPTED for HAaRO */
#define IS_REGREP_ACCEPTED(code) \
	((code) == REGREP_ACCEPTED || (code) == REGREP_ACCEPTED_NO_SB || (code) == REGREP_PREACCEPTED)


/* Registration Reply */
struct reg_rep {
	__u8 type;
	__u8 code;
	__u16 lifetime;
	struct in_addr home_addr;
	struct in_addr ha_addr;
	__u32 id[2];  /* timestamp */
} ATTRIBUTE_PACKED;


/* FA Registration; message types 240 and 241 have not been reserved yet
 * These messages must have extensions FA NAI and FA-FA auth */

#define FA_REG_NONCE_LEN 8

#define FA_REQ 240

#define FA_REQ_NONCE_SYNC 0x80

struct fa_reg_req {
	__u8 type; /* FA_REQ */
	__u8 opts; /* option flags, FA_REQ_* */
	__u16 lifetime; /* registration lifetime in seconds */
	__u8 up_nonce[FA_REG_NONCE_LEN]; /* previous nonce(up) */
	__u8 down_nonce[FA_REG_NONCE_LEN]; /* next nonce(down) */
} ATTRIBUTE_PACKED;

#define FA_REP 241

/* accept */
#define FA_REP_OK 0
/* denial */
#define FA_REP_UNKNOWN_REASON 1
#define FA_REP_NONCE_FAILED 2
#define FA_REP_AUTH_FAILED 3
#define FA_REP_INVALID_REQ 4

struct fa_reg_rep {
	__u8 type; /* FA_REP */
	__u8 code; /* FA_REP_* */
	__u8 up_nonce[FA_REG_NONCE_LEN]; /* next nonce(up) */
	__u8 down_nonce[FA_REG_NONCE_LEN]; /* previous nonce(down) */
} ATTRIBUTE_PACKED;

/*
 * extensions
 */

/* Own vendor extension for registration messages */
struct registration_ext_dynamics {
	VENDOR_EXT_HEADER
	__u8 version; /* vendor extension version */
	__u8 opts;
	__u32 seq; /* MN req. sequence number for local location update
		    * replay protection (increasing for current session key) */
} ATTRIBUTE_PACKED;

#define REG_EXT_OWN_TEAR_DOWN 0x80


/* RFC 2002
 * msg_auth type can be:
 *
 * 32 for Mobile-Home Authentication
 * 33 for Mobile-Foreign Authentication
 * 34 for Foreign-Home Authentication
 *  - contains MAC based on shared secret
 */
#define MH_AUTH 32
#define MF_AUTH 33
#define FH_AUTH 34

#define SPI_LEN 4
#define GET_AUTH_EXT_LEN(msg_key) (msg_key->length + 2)
#define GET_AUTH_LEN(msg_key) (msg_key->length - 4)
#define MSG_AUTH_DATA(msg) (((unsigned char*)msg) + sizeof(struct msg_auth))

struct msg_auth {
	__u8 type;
	__u8 length;
	__u32 spi;
} ATTRIBUTE_PACKED;


/* Dynamics Authentication Additions (as vendor extensions)
 * VENDOR_EXT_DYNAMICS_FF_AUTH 9
 * - Foreign-Foreign Authentication (used like extensions 32 - 34 in RFC 2002,
 *   but between Foreign Agents
 * VENDOR_EXT_DYNAMICS_SK_AUTH 10
 * - Session Key based authentication (i.e. session key is used like shared
 *   secret)
 * - used in local location updates
 * VENDOR_EXT_DYNAMICS_SHA_HA_AUTH 17
 * - Surrogate Home Agent-private Home Agent Authentication
 */

#define GET_VENDOR_AUTH_EXT_LEN(msg_key) (msg_key->length + 2)
#define GET_VENDOR_AUTH_LEN(msg_key) (msg_key->length - (NVSE_HEADER_LEN + 4))
#define MIN_VENDOR_AUTH_LEN (NVSE_HEADER_LEN + 4)
#define MSG_VENDOR_AUTH_DATA(msg) (((unsigned char*)msg) + \
sizeof(struct vendor_msg_auth))

struct vendor_msg_auth {
	VENDOR_EXT_HEADER
	__u32 spi;
} ATTRIBUTE_PACKED;


/* Most of these extensions are based on regkey draft, but the type of the
 * length field is changed (now 8 bits) and the implemention is using vendor
 * extensions with type > 127
 *
 * msg_key vendor sub_type can be:
 *
 * Foreign Agent Key Request (regkey: type=113)
 * VENDOR_EXT_DYNAMICS_FA_KEYREQ 3
 *  - sent by foreign agents to next higher agent if there is a security
 *    association to that agent
 *  - does not have key data (i.e. length = 4 always)
 *  - sent in registration requests
 *
 * Foreign Agent Public Key (regkey: type=115)
 * VENDOR_EXT_DYNAMICS_FA_PUBKEY 4
 *  - sent by foreign agents to next higher agent if no security association
 *    is available
 *  - contains foreign agent public key (RSA)
 *  - sent in registration requests
 *
 * Mobile Node Key Request (regkey: no similar type)
 * VENDOR_EXT_DYNAMICS_MN_KEYREQ 5
 *  - sent by mobile node to home agent
 *  - does not have key data (i.e. length = 4 always)
 *  - sent in registration requests
 *
 * Home-Mobile Key Reply (regkey: type=120)
 * VENDOR_EXT_DYNAMICS_MN_KEYREP 6
 *  - sent by home agent (indirectly) to mobile node
 *  - contains session key encrypted with shared secret
 *  - sent in registration replies
 *
 * Foreign Agent Key Reply (regkey: type=121)
 * VENDOR_EXT_DYNAMICS_FA_KEYREP 7
 *  - sent by agents to next lower foreign agent
 *  - contains session key encrypted with shared secret
 *  - sent in registration replies
 *
 * Foreign Agent Public Key Reply (regkey: type=123)
 * VENDOR_EXT_DYNAMICS_FA_PUBKEYREP 8
 *  - sent by agents to next lower foreign agent
 *  - contains encrypted session key (RSA)
 *  - sent in registration replies
 *
 * Public Key Hash
 * VENDOR_EXT_DYNAMICS_PUBKEY_HASH 11
 *  - sent by FAs in the agentadv and by MN in the registration request
 *  - used for man-in-the-middle style attack prevention
 *  - contains the hash code of the HFA's public key
 */

#define GET_KEY_EXT_LEN(msg_key) (msg_key->length + 2)
#define GET_KEY_LEN(msg_key) (msg_key->length - (NVSE_HEADER_LEN + 4))
#define MIN_KEY_EXT_LEN (NVSE_HEADER_LEN + 4)
#define MSG_KEY_DATA(msg) ((unsigned char*)(((char*)msg) + \
        sizeof(struct msg_key)))

#define PUBKEY_ALG_SPI_RSA 1000

struct msg_key {
	VENDOR_EXT_HEADER
	__u32 spi;
} ATTRIBUTE_PACKED;

/* See draft-ietf-mobileip-regkey-00.txt for more detailed description
 * of the key request/reply extensions.
 */

/* VENDOR_EXT_DYNAMICS_FA_NAI 12 */
/* VENDOR_EXT_DYNAMICS_PREVIOUS_FA_NAI 13 */
#define GET_NAI_EXT_LEN(msg_nai) (msg_nai->length + 2)
#define GET_NAI_LEN(msg_nai) (msg_nai->length - NVSE_HEADER_LEN)
#define MSG_NAI_DATA(msg_nai) ((unsigned char*)(((char*)msg_nai) + \
        sizeof(struct fa_nai_ext)))

#ifndef MAX_NAI_LEN
#define MAX_NAI_LEN 128
#endif

struct fa_nai_ext {
	VENDOR_EXT_HEADER
} ATTRIBUTE_PACKED;


/* VENDOR_EXT_DYNAMICS_GRE_KEY 14 */
#define GET_GRE_KEY_EXT_LEN(msg_gre) (msg_gre->length + 2)
#define GET_GRE_KEY_LEN(msg_gre) (msg_gre->length - NVSE_HEADER_LEN)
struct gre_key_ext {
	VENDOR_EXT_HEADER
	__u32 key;
} ATTRIBUTE_PACKED;

/* VENDOR_EXT_DYNAMICS_SFA_DEBUG 15 */
#define GET_SFA_DEBUG_EXT_LEN(msg_sfa) (msg_sfa->length + 2)
struct sfa_debug_ext {
	VENDOR_EXT_HEADER
	struct in_addr sfa_addr;
} ATTRIBUTE_PACKED;

/* VENDOR_EXT_DYNAMICS_PRIV_HA 16 */
#define GET_PRIV_HA_EXT_LEN(msg) (msg->length + 2)
struct priv_ha_ext {
	VENDOR_EXT_HEADER
	__u32 priv_ha;
} ATTRIBUTE_PACKED;

/* VENDOR_EXT_DYNAMICS_NONCE 18 */
#define GET_NONCE_EXT_LEN(msg) (msg->length + 2)
struct nonce_ext {
	VENDOR_EXT_HEADER
	__u32 nonce;
} ATTRIBUTE_PACKED;

/* RFC 2002 */
#define ONE_BYTE_PADDING 0


/* RFC 3024 */

#define ENCAPS_DELIVERY_EXT 130
struct encaps_delivery_ext {
	__u8 type; /* 130 */
	__u8 length; /* 0 */
} ATTRIBUTE_PACKED;


/* RFC 2794 */

#define MN_NAI_EXT 131

#define GET_MN_NAI_EXT_LEN(msg_nai) (msg_nai->length + 2)
#define GET_MN_NAI_LEN(msg_nai) (msg_nai->length)
#define MSG_MN_NAI_DATA(msg_nai) ((unsigned char*)(((char*)msg_nai) + \
        sizeof(struct mn_nai_ext)))

struct mn_nai_ext {
	__u8 type; /* 131 */
	__u8 length;
} ATTRIBUTE_PACKED;


/* RFC 3012 */

#define AGENT_ADV_CHALLENGE_EXT 24
#define MN_FA_CHALLENGE_EXT 132

#define GET_CHALLENGE_EXT_LEN(ext) (ext->length + 2)
#define GET_CHALLENGE_LEN(ext) (ext->length)
#define MSG_CHALLENGE_EXT_DATA(ext) ((unsigned char *)(((char *) ext) + \
        sizeof(struct challenge_ext)))
#define MAX_CHALLENGE_LEN 256

struct challenge_ext {
	__u8 type; /* 24 / 132 */
	__u8 length;
	/* followed by 0 .. 256 bytes of challenge data */
} ATTRIBUTE_PACKED;


#define GENERALIZED_AUTH_EXT 36

/* generalized auth. ext. subtypes */
#define GENERALIZED_AUTH_MN_AAA 1

#define GET_GEN_AUTH_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_AUTH_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_AUTH_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_auth_ext))

struct generalized_auth_ext {
	__u8 type; /* 36 */
	__u8 subtype; /* 1 */
	__u16 length;
	__u32 spi;
	/* followed by 16 or more bytes of authenticator data */
} ATTRIBUTE_PACKED;

#define CHAP_SPI 2


/* draft-ietf-mobileip-gen-key-00 */

/* Generalized MN-FA Key Request Extension */

#define GENERALIZED_MN_FA_KEY_REQ_EXT 40

#define GET_GEN_MN_FA_KEY_REQ_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_MN_FA_KEY_REQ_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_MN_FA_KEY_REQ_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_mn_fa_key_req_ext))

struct generalized_mn_fa_key_req_ext {
	__u8 type; /* 40 */
	__u8 subtype;
	__u16 length;
	__u32 mn_spi;
	/* followed by (SHOULD be at least 16 bytes of) subtype data */
} ATTRIBUTE_PACKED;

/* MN-FA Key Request subtypes */
/* draft-ietf-mobileip-aaa-key-07.txt */
#define GEN_MN_FA_KEY_REQ_FROM_AAA 7
/* this extension MUST appear before MN-AAA auth. ext. */


/* Generalized MN-FA Key Reply Extension */

#define GENERALIZED_MN_FA_KEY_REP_EXT 41

#define GET_GEN_MN_FA_KEY_REP_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_MN_FA_KEY_REP_LEN(ext) (ntohs(ext->length))
#define MSG_GEN_MN_FA_KEY_REP_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_mn_fa_key_rep_ext))

struct generalized_mn_fa_key_rep_ext {
	__u8 type; /* 41 */
	__u8 subtype;
	__u16 length;
	/* followed by subtype data */
} ATTRIBUTE_PACKED;

/* MN-FA Key Reply subtypes */
/* draft-ietf-mobileip-aaa-key-07.txt */
#define GEN_MN_FA_KEY_REP_KEY_MATERIAL_FROM_AAA 7
/* this extension MUST appear before MN-FA auth. ext. */
struct mn_fa_key_material_from_aaa {
	u32 lifetime;
	u32 aaa_spi;
	u32 fa_spi;
	u16 alg_id;
	/* followed by key material; a random value of at least 64 bits */
} ATTRIBUTE_PACKED;


/* Generalized MN-HA Key Request Extension */

#define GENERALIZED_MN_HA_KEY_REQ_EXT 42

#define GET_GEN_MN_HA_KEY_REQ_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_MN_HA_KEY_REQ_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_MN_HA_KEY_REQ_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_mn_ha_key_req_ext))

struct generalized_mn_ha_key_req_ext {
	__u8 type; /* 42 */
	__u8 subtype;
	__u16 length;
	__u32 mn_spi;
	/* followed by (SHOULD be at least 16 bytes of) subtype data */
} ATTRIBUTE_PACKED;

/* MN-HA Key Request subtypes */
/* draft-ietf-mobileip-aaa-key-07.txt */
#define GEN_MN_HA_KEY_REQ_FROM_AAA 7
/* this extension MUST appear before MN-AAA auth. ext. */


/* Generalized MN-HA Key Reply Extension */

#define GENERALIZED_MN_HA_KEY_REP_EXT 43

#define GET_GEN_MN_HA_KEY_REP_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_MN_HA_KEY_REP_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_MN_HA_KEY_REP_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_mn_ha_key_rep_ext))

struct generalized_mn_ha_key_rep_ext {
	__u8 type; /* 43 */
	__u8 subtype;
	__u16 length;
	__u32 lifetime;
	/* followed by subtype data */
} ATTRIBUTE_PACKED;

/* MN-HA Key Reply subtypes */
/* draft-ietf-mobileip-aaa-key-07.txt */
#define GEN_MN_HA_KEY_REP_KEY_MATERIAL_FROM_AAA 1
/* this extension MUST appear before MN-HA auth. ext. */
struct mn_ha_key_material_from_aaa {
	u32 aaa_spi;
	u32 ha_spi;
	u16 alg_id;
	u16 replay_method;
	/* followed by key material; a random value of at least 64 bits */
} ATTRIBUTE_PACKED;


/* Generalized FA-HA Key Reply Extension */

#define GENERALIZED_FA_HA_KEY_REP_EXT 45

#define GET_GEN_FA_HA_KEY_REP_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_FA_HA_KEY_REP_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_FA_HA_KEY_REP_DATA(ext) (((unsigned char *) ext) + \
	sizeof(struct generalized_fa_ha_key_rep_ext))

struct generalized_fa_ha_key_rep_ext {
	__u8 type; /* 45 */
	__u8 subtype;
	__u16 length;
	__u32 lifetime;
	/* followed by subtype data */
} ATTRIBUTE_PACKED;


/* Generalized FA-FA Key Reply Extension */

#define GENERALIZED_FA_FA_KEY_REP_EXT 46

#define GET_GEN_FA_FA_KEY_REP_EXT_LEN(ext) (ntohs(ext->length) + 4)
#define GET_GEN_FA_FA_KEY_REP_LEN(ext) (ntohs(ext->length) - 4)
#define MSG_GEN_FA_FA_KEY_REP_DATA(ext) (((unsigned cfar *) ext) + \
	sizeof(struct generalized_fa_fa_key_rep_ext))

struct generalized_fa_fa_key_rep_ext {
	__u8 type; /* 46 */
	__u8 subtype;
	__u16 length;
	__u32 fa_fa_spi;
	/* followed by subtype data */
} ATTRIBUTE_PACKED;


/* draft-ietf-mobileip-aaa-key-07.txt */

/* Algorithm Identifiers */
#define AAA_KEY_ALG_MD5 1		/* MD5/prefix+suffix, RFC 2002 */
#define AAA_KEY_ALG_HMAC_MD5 2		/* HMAC MD5, RFC 2104 */
#define AAA_KEY_ALG_SHA1 3		/* FIPS 180-1 */

/* Replay Methods */
#define AAA_KEY_REPLAY_NONE 1		/* RFC 2002 */
#define AAA_KEY_REPLAY_TIMESTAMPS 2	/* RFC 2002 */
#define AAA_KEY_REPLAY_NONCES 3		/* RFC 2002 */


/*--------------------------------------------------------------*
 | jpaaso - UDP Tunneling messages.								|
 *--------------------------------------------------------------*/

/* MIP msg type 4: UDP Tunnel Message */
#define TNL 4

/* Registration Extension types */
#define UDP_TNL_REQ_EXT 144
#define UDP_TNL_REP_EXT 44

/* UDP Tunnel Request Extension */
struct udp_tnl_req_ext {
	__u8 type;			// == 144	
	__u8 length;		// == 6
	__u8 subtype;		// == 0
	__u8 reserved1;		// == 0
	__u8 reserved2;		// == 0
	__u8 encapsulation;
	__u16 reserved3;	// == 0
} ATTRIBUTE_PACKED; 

/* UDP Tunnel Reply Extension */
struct udp_tnl_rep_ext {
	__u8 type;			// == 44	
	__u8 length;		// == 6
	__u8 subtype;		// == 0
	__u8 reply_code;	// tunneling accepted / not accepted
	__u16 reserved;
	__u16 keepalive_interval;
} ATTRIBUTE_PACKED; 

/* MIP Tunnel Data Message */
struct udp_tnl_data {
	__u8 type;			// == 4	
	__u8 next_header;	// == 4
	__u16 reserved;		// == 0 when sending. Ignored when receiving.
} ATTRIBUTE_PACKED; 


/*--------------------------------------------------------------*
 | jpaaso - Network Mobility (NEMO) extensions.					|
 *--------------------------------------------------------------*/

#define MOBILE_NW_EXT 148
#define NEMO_NW_MAX_AMOUNT 16	/* MN allowed to have 16 networks max */
#define NEMO_REQ_SUBTYPE 0 
#define NEMO_EXP_ACK_SUBTYPE 1
#define NEMO_IMP_ACK_SUBTYPE 2

struct mobile_nw_req_ext {
	__u8 type;			// == 148
	__u8 length;		// == 6
	__u8 subtype;		// == 0
	__u8 prefix_length;	// 0 - 32
	__u32 prefix;		// IPv4 address in nw byte order (s_addr)
} ATTRIBUTE_PACKED; 

/* Mobile NW Ack Extension codes */
#define MOBNET_SUCCESS 0 				// success 
#define MOBNET_INVALID_PREFIX_LEN 1 	// invalid prefix
#define MOBNET_UNAUTHORIZED 2 			// MR not authorized for prefix
#define MOBNET_FWDING_SETUP_FAILED 3	// forwarding setup failed

struct mobile_nw_ack_ext {
	__u8 type;			// == 148
	__u8 length;		// == 8
	__u8 subtype;		// 1 == explicit mode ack, 2 == implicit mode ack
	__u8 code;			
	__u8 prefix_length;	// 0 - 32
	__u8 reserved;		// sent as 0, ignored on reception
	__u32 prefix;		// IPv4 address in nw byte order
} ATTRIBUTE_PACKED; 



/*--------------------------------------------------------------*
 | jpaaso - Route Optimization (HAaRO) extensions.				|
 *--------------------------------------------------------------*/
/* HAaRO temporary types randomly chosen */
#define TBA_T1 254
#define TBA_T2 126
#define TBA_T3 125

/* HAaRO temporary subtypes randomly chosen */
#define TBA_ST1_1 124
#define TBA_ST2_1 123
#define TBA_ST_2_2 122
#define TBA_ST_2_3 121
#define TBA_ST3_1 120

/* New MIP message types */
#define TBA_MIP1 111	/* HoTI */
#define TBA_MIP2 112	/* CoTI */
#define TBA_MIP3 113	/* HoT 	*/
#define TBA_MIP4 114	/* CoT 	*/

/* RO Reply codes */
#define RO_OK 0
#define RO_DECLINED 16

/* RO Capability flags */
#define RO_CAPAB_A 0x80
#define RO_CAPAB_R 0x40
#define RO_CAPAB_S 0x20
#define RO_CAPAB_O 0x10

/* RO Reply flags */
#define RO_REPLY_O 0x80
#define RO_REPLY_N 0x40
#define RO_REPLY_S 0x20

/* RO Prefix Advertisement flags */
#define RO_ADV_D 0x80
#define RO_ADV_M 0x40

/* S flag set -> advertise only networks of the MR whose HoA
 * is placed in this packet */
struct ro_capab_s_on {
	__u8 type;				// == TBA_T1
	__u8 length;			// == 6
	__u8 subtype;			// == TBA_ST1_1
	__u8 flags;				// |A|R|S|O|  Rsvd |
	__u32 mr_hoa;			// IPv4 address in nw byte order
} ATTRIBUTE_PACKED;

/* S flag not set -> reply with all known MR NEMO nws */
struct ro_capab_s_off {
	__u8 type;				// == TBA_T1
	__u8 length;			// == 2
	__u8 subtype;			// == TBA_ST1_1
	__u8 flags;				// |A|R|S|O|  Rsvd |
} ATTRIBUTE_PACKED;

/* MR Route Optimization Capability */
struct ro_capab_ext {
	struct ro_capab_s_on 	*s_on;
	struct ro_capab_s_off 	*s_off;
} ATTRIBUTE_PACKED;

/* Route Optimization Reply */
struct ro_reply_ext {
	__u8 type;					// == TBA_T2
	__u8 length;				// == 2 
	__u8 subtype;				// == TBA_ST2_1
	__u8 code;					// 0 == RO ok, 16 == RO declined 
} ATTRIBUTE_PACKED;

/* Mobile-Correspondent Authentication Extension */ 
struct mr_cr_auth_ext {
	__u8 type;					// == TBA_T2
	__u8 length;				// == 22
	__u8 subtype;				// == TBA_ST_2_2
	__u8 reserved;
	__u16 home_nonce_index;
	__u16 care_of_nonce_index;
	__u64 authenticator[2];		// = HMAC_MD5(KRm, (The whole Reg packet until this field))
} ATTRIBUTE_PACKED;

/* Care-of Address Extension */
struct coa_ext {
	__u8 type;					// == TBA_T2
	__u8 length;				// == 2 + 4 * n
	__u8 subtype;				// == TBA_ST_2_3
	__u8 reserved; 			
	// 1-n times the following
	__u32 coa;					// IPv4 address in nw byte order
} ATTRIBUTE_PACKED;

struct ro_prefix_adv_head {
	__u8 type;					// == TBA_T3
	__u8 subtype;				// == TBA_ST_2_3
	__u16 length;				// == 1
} ATTRIBUTE_PACKED;

/* Data structure appears 1-n times after the head. One data item
 * contains one of the following depending on the M bit's state:
 *  - HoA of a RO capable MR (M == 1)
 *  - Prefix + Realm of a NEMO network behind this MR (M == 0)
 * 
 * [ MR1 HoA | MR1 NEMO nw1 | MR1 NEMO nw1 NAI | MR1 NEMO nw2 | ...
 *   MR2 HoA | MR2 NEMO nw1 | MR2 NEMO nw1 NAI | MR2 NEMO nw2 | ...
 * 
 */
struct ro_prefix_adv_data {
	__u8 flags_plen_info;		// |D|M| Plen/Info |
	__u32 hoa_or_prefix;		// HoA: 4 B or Prefix (1, 2, 3 or) 4 B
	/* Prefix is followed by a realm (1-n characters) */
} ATTRIBUTE_PACKED;

/* Route Optimization Prefix Advertisement 
 *  - head->length contains the total length AFTER the length field
 *  - data pointer can be used to iterate through data items 
 */
struct ro_prefix_adv {
	struct ro_prefix_adv_head *head;
	struct ro_prefix_adv_data *data;
} ATTRIBUTE_PACKED;

/* Home Test Init message */
struct ro_hoti {
	__u8 type;		// TBA_MIP1
	__u8 reserved;	// 0
	__u64 cookie;	// rand()
} ATTRIBUTE_PACKED;

/* Care-of-Test Init message */
struct ro_coti {
	__u8 type;		// TBA_MIP2
	__u8 reserved;	// 0
	__u64 cookie;	// rand()
} ATTRIBUTE_PACKED;

/* Home Test message */
struct ro_hot {
	__u8 type;			// TBA_MIP3
	__u8 reserved;		// 0
	__u16 nonce_index;	// echoed back MR->CR in RegReq auth ext
	__u64 cookie;		// = hoti.cookie
	__u64 keygen_token;	// generated from cookie + nonce
} ATTRIBUTE_PACKED;

/* Care-of-Test message */
struct ro_cot {
	__u8 type;			// TBA_MIP4
	__u8 reserved;		// 0
	__u16 nonce_index;	// echoed back MR->CR in RegReq auth ext
	__u64 cookie;		// = coti.cookie
	__u64 keygen_token;	// generated from cookie + nonce
} ATTRIBUTE_PACKED;

#endif /* MESSAGE_H */
